#
# Copyright 2015-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Arc Calibration Recipe for Megara"""

from __future__ import division, print_function

import traceback
from datetime import datetime

import numpy
from astropy.io import fits
from copy import deepcopy
import errno
import os

from numina.core import Requirement, Product, Parameter, DataFrameType
from numina.core.requirements import ObservationResultRequirement
from numina.array.display.polfit_residuals import polfit_residuals
from numina.array.display.polfit_residuals import \
    polfit_residuals_with_sigma_rejection
from numina.array.display.ximplotxy import ximplotxy
from numina.array.wavecalib.__main__ import find_fxpeaks
from numina.array.wavecalib.arccalibration import arccalibration_direct
from numina.array.wavecalib.arccalibration import fit_list_of_wvfeatures
from numina.array.wavecalib.arccalibration import gen_triplets_master
from numina.array.wavecalib.arccalibration import refine_arccalibration
from numina.array.wavecalib.solutionarc import CrLinear
from numina.array.wavecalib.solutionarc import SolutionArcCalibration
from numina.core.validator import range_validator
from numina.util.flow import SerialFlow
from numina.array import combine

from megaradrp.types import ProcessedFrame, ProcessedRSS
from megaradrp.processing.combine import basic_processing_with_combination
from megaradrp.processing.aperture import ApertureExtractor
from megaradrp.processing.fiberflat import Splitter, FlipLR
from megaradrp.core.recipe import MegaraBaseRecipe
from megaradrp.products import WavelengthCalibration
from megaradrp.products.wavecalibration import FiberSolutionArcCalibration
import megaradrp.requirements as reqs

from megaradrp.instrument import vph_thr_arc


class ArcCalibrationRecipe(MegaraBaseRecipe):
    """Provides wavelength calibration information from arc images.

    This recipes process a set of arc images obtained in
    *Arc Calibration* mode and returns the information required
    to perform wavelength calibration and resampling in other recipes,
    in the form of a WavelengthCalibration object. The recipe also returns
    a 2D map of the FWHM of the arc lines used for the calibration,
    the result images up to dark correction, and the result of the processing
    up to aperture extraction.


    See Also
    --------
    megaradrp.products.WavelengthCalibration: description of WavelengthCalibration product
    megaradrp.products.LinesCatalog: description of the catalog of lines
    megaradrp.processing.aperture: aperture extraction
    numina.array.wavecalib.arccalibration: wavelength calibration algorithm
    megaradrp.instrument.configs: instrument configuration

    Notes
    -----
    Images provided in `obresult` are trimmed and corrected from overscan,
    bad pixel mask (if `master_bpm` is not None), bias and dark current
    (if `master_dark` is not None).
    Images thus corrected are the stacked using the median.

    The result of the combination is saved as an intermediate result, named
    'reduced_image.fits'. This combined image is also returned in the field
    `reduced_image` of the recipe result.

    The apertures in the 2D image are extracted, using the information in
    `master_traces`. the result of the extraction is saved as an intermediate
    result named 'reduced_rss.fits'. This RSS is also returned in the field
    `reduced_rss` of the recipe result.

    For each fiber in the reduced RSS, the peaks are detected and sorted
    by peak flux. `nlines` is used to select the brightest peaks. If it is a
    list, then the peaks are divided, by their position, in as many groups as
    elements in the list and `nlines[0]` peaks are selected in the first
    group, `nlines[1]` peaks in the second, etc.

    The selected peaks are matched against the catalog of lines in `lines_catalog`.
    The matched lines, the quality of the match and other relevant information is
    stored in the product WavelengthCalibration object. The wavelength of the matched
    features is fitted to a polynomial
    of degree `polynomial_degree`. The coefficients of the polynomial are
    stored in the final `master_wlcalib` object for each fiber.

    """

    # Requirements
    obresult = ObservationResultRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_bpm = reqs.MasterBPMRequirement()
    master_apertures = reqs.MasterAperturesRequirement()
    extraction_offset = Parameter([0.0], 'Offset traces for extraction',
                                  accept_scalar=True)
    lines_catalog = reqs.LinesCatalogRequirement()
    polynomial_degree = Parameter(5, 'Polynomial degree of arc calibration',
                                  as_list=True, nelem='+',
                                  validator=range_validator(minval=1)
                                  )
    nlines = Parameter(20, "Use the 'nlines' brigthest lines of the spectrum",
                       as_list=True, nelem='+',
                       validator=range_validator(minval=0))
    debug_plot = Parameter(0, 'Save intermediate tracing plots')
    store_pdf_with_refined_fits = Parameter(
        0,
        description='Store PDF plot with refined fits for each fiber',
    )

    # Products
    reduced_image = Product(ProcessedFrame)
    reduced_rss = Product(ProcessedRSS)
    master_wlcalib = Product(WavelengthCalibration)
    fwhm_image = Product(DataFrameType)

    def run(self, rinput):
        """Execute the recipe.

        Parameters
        ----------
        rinput : ArcCalibrationRecipe.RecipeInput

        Returns
        -------
        ArcCalibrationRecipe.RecipeResult

        """

        self.logger.info('starting arc calibration recipe')

        debugplot = rinput.debug_plot if self.intermediate_results else 0

        obresult = rinput.obresult
        obresult_meta = obresult.metadata_with(self.datamodel)

        flow1 = self.init_filters(rinput, rinput.obresult.configuration)
        img = basic_processing_with_combination(rinput, flow1, method=combine.median)
        hdr = img[0].header
        self.set_base_headers(hdr)

        self.save_intermediate_img(img, 'reduced_image.fits')

        splitter1 = Splitter()
        calibrator_aper = ApertureExtractor(
            rinput.master_apertures,
            self.datamodel,
            offset=rinput.extraction_offset
        )
        flipcor = FlipLR()

        flow2 = SerialFlow([splitter1, calibrator_aper, flipcor])

        reduced_rss = flow2(img)
        self.save_intermediate_img(reduced_rss, 'reduced_rss.fits')

        reduced2d = splitter1.out

        self.logger.info('extract fibers, %i', len(rinput.master_apertures.contents))

        current_vph = rinput.obresult.tags['vph']
        current_insmode = rinput.obresult.tags['insmode']

        if current_insmode in vph_thr_arc and current_vph in vph_thr_arc[current_insmode]:
            threshold = vph_thr_arc[current_insmode][current_vph]['threshold']
            min_distance = vph_thr_arc[current_insmode][current_vph]['min_distance']
            self.logger.info('rel threshold for %s is %4.2f', current_vph, threshold)
        else:
            threshold = 0.02
            min_distance = 10.0
            self.logger.info('rel threshold not defined for %s, using %4.2f', current_vph, threshold)
            self.logger.info('min dist not defined for %s, using %4.2f', current_vph, min_distance)

        # WL calibration goes here
        initial_data_wlcalib, data_wlcalib, fwhm_image = self.calibrate_wl(
            reduced_rss[0].data,
            rinput.lines_catalog,
            rinput.polynomial_degree,
            rinput.master_apertures, rinput.nlines,
            threshold=threshold,
            min_distance=min_distance,
            debugplot=debugplot,
            store_pdf_with_refined_fits=rinput.store_pdf_with_refined_fits
        )

        initial_data_wlcalib.tags = rinput.obresult.tags
        final = initial_data_wlcalib
        final.update_metadata(self)
        final.update_metadata_origin(obresult_meta)

        self.save_structured_as_json(
            initial_data_wlcalib,
            'initial_master_wlcalib.json'
        )

        self.logger.info('end arc calibration recipe')

        if data_wlcalib is None:
            return self.create_result(
                reduced_image=reduced2d,
                reduced_rss=reduced_rss,
                master_wlcalib=initial_data_wlcalib,
                fwhm_image=fwhm_image
            )
        else:
            # copy metadata from initial_master_wlcalib to master_wlcalib
            data_wlcalib.tags = deepcopy(initial_data_wlcalib.tags)
            data_wlcalib.meta_info = deepcopy(initial_data_wlcalib.meta_info)
            return self.create_result(
                reduced_image=reduced2d,
                reduced_rss=reduced_rss,
                master_wlcalib=data_wlcalib,
                fwhm_image=fwhm_image
            )

    def calc_fwhm_of_line(self, row, peak_int, lwidth=20):
        """
        Compute FWHM of lines in spectra
        """
        import numina.array.fwhm as fmod

        # FIXME: this could wrap around the image
        qslit = row[peak_int - lwidth:peak_int + lwidth]
        return fmod.compute_fwhm_1d_simple(qslit, lwidth)

    def calibrate_wl(self, rss, lines_catalog, poldeg, tracemap, nlines,
                     threshold=0.27,
                     min_distance=30,
                     debugplot=0,
                     store_pdf_with_refined_fits=0):

        if len(poldeg) == 1:
            poldeg_initial = poldeg[0]
            poldeg_refined = poldeg[0]
        elif len(poldeg) == 2:
            poldeg_initial = poldeg[0]
            poldeg_refined = poldeg[1]
        else:
            raise ValueError("Unexpected list length for polynomial_degree")

        if poldeg_initial > poldeg_refined:
            raise ValueError("Unexpected poldeg_initial (" +
                             str(poldeg_initial) + ") > poldeg_refined (" +
                             str(poldeg_refined) + ")")

        wv_master_all = lines_catalog[:, 0]
        if lines_catalog.shape[1] == 2:  # assume old format
            wv_master = numpy.copy(wv_master_all)
            refine_wv_calibration = False
            if abs(debugplot) >= 10:
                print('wv_master:\n', wv_master)
        elif lines_catalog.shape[1] == 3:  # assume new format
            wv_flag = lines_catalog[:, 1]
            wv_master = wv_master_all[numpy.where(wv_flag == 1)]
            refine_wv_calibration = True
            if abs(debugplot) >= 10:
                print('wv_master:\n', wv_master)
        else:
            raise ValueError('lines_catalog file does not have the expected '
                             'number of columns')

        # minimum and maximum wavelength
        wv_range_catalog = wv_master_all[-1] - wv_master_all[0]
        delta_wv = 0.20 * wv_range_catalog
        wv_ini_search = int(wv_master_all[0] - delta_wv)
        wv_end_search = int(wv_master_all[-1] + delta_wv)
        self.logger.info('wv_ini_search %s', wv_ini_search)
        self.logger.info('wv_end_search %s', wv_end_search)

        ntriplets_master, ratios_master_sorted, triplets_master_sorted_list = \
            gen_triplets_master(wv_master)

        error_contador = 0
        missing_fib = 0
        crpix1 = 1.0
        naxis1 = rss.shape[1]

        plot_tracenumber = []
        plot_npeaksfound = []
        plot_crval1 = []
        plot_cdelt1 = []
        plot_coeff = []

        initial_data_wlcalib = WavelengthCalibration(instrument='MEGARA')
        initial_data_wlcalib.total_fibers = tracemap.total_fibers
        for trace in tracemap.contents:
            fibid = trace.fibid
            idx = trace.fibid - 1

            if trace.valid:
                self.logger.info('-' * 52)
                self.logger.info('Starting row %d, fibid %d', idx, fibid)

                row = rss[idx]

                fxpeaks, sxpeaks = find_fxpeaks(
                    sp=row,
                    times_sigma_threshold=0.0,
                    minimum_threshold=0,
                    nwinwidth_initial=7,
                    nwinwidth_refined=5,
                    npix_avoid_border=6,
                    nbrightlines=nlines,
                    sigma_gaussian_filtering=0,
                    minimum_gaussian_filtering=0
                )
                self.logger.info('number of peaks (expected): %s',
                                 str(nlines))
                self.logger.info('number of peaks (found)...: %d',
                                 len(fxpeaks))

                try:

                    # use channels (pixels from 1 to naxis1)
                    xchannel = fxpeaks + 1.0

                    list_of_wvfeatures = arccalibration_direct(
                        wv_master=wv_master,
                        ntriplets_master=ntriplets_master,
                        ratios_master_sorted=ratios_master_sorted,
                        triplets_master_sorted_list=triplets_master_sorted_list,
                        xpos_arc=xchannel,
                        naxis1_arc=naxis1,
                        crpix1=crpix1,
                        wv_ini_search=wv_ini_search,
                        wv_end_search=wv_end_search,
                        error_xpos_arc=3.0, # initially: 2.0
                        times_sigma_r=3.0,
                        frac_triplets_for_sum=0.50,
                        times_sigma_theil_sen=10.0,
                        poly_degree_wfit=poldeg_initial,
                        times_sigma_polfilt=10.0,
                        times_sigma_cook=10.0,
                        times_sigma_inclusion=10.0,
                        debugplot=debugplot
                    )

                    self.logger.info('Solution for row %d completed', idx)
                    self.logger.info('Fitting solution for row %d', idx)
                    solution_wv = fit_list_of_wvfeatures(
                            list_of_wvfeatures,
                            naxis1_arc=naxis1,
                            crpix1=crpix1,
                            poly_degree_wfit=poldeg_initial,
                            weighted=False,
                            debugplot=0,
                            plot_title=None
                        )

                    self.logger.info('linear crval1, cdelt1: %f %f',
                                     solution_wv.cr_linear.crval,
                                     solution_wv.cr_linear.cdelt)

                    self.logger.info('fitted coefficients %s',
                                     solution_wv.coeff)

                    # store results for plotting
                    plot_tracenumber.append(fibid)
                    plot_npeaksfound.append(len(fxpeaks))
                    plot_crval1.append(solution_wv.cr_linear.crval)
                    plot_cdelt1.append(solution_wv.cr_linear.cdelt)
                    plot_coeff.append(solution_wv.coeff)

                    trace_pol = trace.polynomial
                    # Update feature with measurements of Y coord in original
                    # image
                    # Peak and FWHM in RSS
                    for feature in solution_wv.features:
                        # Compute Y
                        feature.ypos = trace_pol(feature.xpos)
                        # FIXME: check here FITS vs PYTHON coordinates, etc
                        peak_int = int(feature.xpos)
                        try:
                            peak, fwhm = self.calc_fwhm_of_line(row, peak_int,
                                                                lwidth=20)
                        except Exception as error:
                            self.logger.error("%s", error)
                            self.logger.error('error in feature %s', feature)
                            # workaround
                            peak = row[peak_int]
                            fwhm = 0.0
                        # I would call this peak instead...
                        feature.peak = peak
                        feature.fwhm = fwhm

                    new = FiberSolutionArcCalibration(fibid, solution_wv)
                    initial_data_wlcalib.contents.append(new)

                except (ValueError, TypeError, IndexError) as error:
                    self.logger.error("%s", error)
                    self.logger.error('error in row %d, fibid %d', idx, fibid)
                    traceback.print_exc()
                    initial_data_wlcalib.error_fitting.append(fibid)
                    error_contador += 1

            else:
                self.logger.info('skipping row %d, fibid %d, not extracted',
                                 idx, fibid)
                missing_fib += 1
                initial_data_wlcalib.missing_fibers.append(fibid)

        self.logger.info('Errors in fitting: %s', error_contador)
        self.logger.info('Missing fibers: %s', missing_fib)

        # save PDF file with plots in working directory
        if self.intermediate_results:
            from numina.array.display.matplotlib_qt import plt
            from matplotlib.backends.backend_pdf import PdfPages
            pdf = PdfPages('wavecal_iter1.pdf')
            for dumplot in zip([plot_npeaksfound, plot_crval1, plot_cdelt1],
                               ['number of peaks found',
                                'linear CRVAL1 ' + r'($\AA$)',
                                'linear_CDELT1 ' + r'($\AA$/pixel)']):
                ax = ximplotxy(plot_tracenumber, dumplot[0],
                               xlabel='fiber number', ylabel=dumplot[1],
                               linestyle='', marker='.', color='C0',
                               show=False)
                pdf.savefig()
                plt.close()
            for ideg in range(poldeg_initial + 1):
                dumplot = [coef[ideg] for coef in plot_coeff]
                ax = ximplotxy(plot_tracenumber, dumplot,
                               xlabel='fiber number',
                               ylabel='coef[' + str(ideg) + ']',
                               linestyle='', marker='.', color='C0',
                               show=False)
                pdf.savefig()
                plt.close()
            pdf.close()

        self.logger.info('Generating fwhm_image...')
        image = self.generate_fwhm_image(initial_data_wlcalib.contents)
        fwhm_image = fits.PrimaryHDU(image)
        fwhm_hdulist = fits.HDUList([fwhm_image])

        if refine_wv_calibration:
            self.logger.info('Improving wavelength calibration...')
            # model polynomial coefficients vs. fiber number using
            # previous results stored in data_wlcalib
            list_poly_vs_fiber = self.model_coeff_vs_fiber(
                initial_data_wlcalib, poldeg_initial,
                times_sigma_reject=5,
                debugplot=0)
            # recompute data_wlcalib from scratch
            missing_fib = 0
            error_contador = 0
            data_wlcalib = WavelengthCalibration(instrument='MEGARA')
            data_wlcalib.total_fibers = tracemap.total_fibers
            plot_tracenumber = []
            plot_npointseff = []
            plot_residualstd = []
            plot_crval1 = []
            plot_cdelt1 = []
            plot_coeff = []

            # output PDF with refined fits
            if store_pdf_with_refined_fits == 1:
                if self.intermediate_results:
                    if not os.path.exists('refined_wavecal'):
                        try:
                            os.makedirs('refined_wavecal')
                        except OSError as exc:  # Guard against race condition
                            if exc.errno != errno.EEXIST:
                                raise
            # refine wavelength calibration polynomial for each valid fiber
            for trace in tracemap.contents:
                fibid = trace.fibid
                idx = trace.fibid - 1
                if trace.valid:
                    self.logger.info('-' * 52)
                    self.logger.info('Starting row %d, fibid %d', idx, fibid)
                    # select spectrum for current fiber
                    row = rss[idx]
                    naxis1 = row.shape[0]
                    # estimate polynomial coefficients for current fiber
                    coeff = numpy.zeros(poldeg_initial + 1)
                    for k in range(poldeg_initial + 1):
                        dumpol = list_poly_vs_fiber[k]
                        coeff[k] = dumpol(fibid)
                    wlpol = numpy.polynomial.Polynomial(coeff)
                    # refine polynomial fit using the full set of arc lines
                    # in master list (and save output PDF file when requested)
                    if store_pdf_with_refined_fits == 1:
                        if self.intermediate_results:
                            from matplotlib.backends.backend_pdf import PdfPages
                            plottitle = 'fiber #{0:03d}'.format(fibid)
                            pdf = PdfPages(
                                'refined_wavecal/{0:03d}.pdf'.format(fibid)
                            )
                        else:
                            plottitle = None
                            pdf = None
                    else:
                        plottitle = None
                        pdf = None
                    poly_refined, yres_summary  = \
                        refine_arccalibration(sp=row,
                                              poly_initial=wlpol,
                                              wv_master=wv_master_all,
                                              poldeg=poldeg_refined,
                                              plottitle=plottitle,
                                              ylogscale=True,
                                              pdf=pdf)
                    if pdf is not None:
                        from numina.array.display.matplotlib_qt import plt
                        plt.close()
                        pdf.close()
                    if poly_refined != numpy.polynomial.Polynomial([0.0]):
                        npoints_eff = yres_summary['npoints']
                        residual_std = yres_summary['robust_std']
                        # compute approximate linear values
                        crmin1_linear = poly_refined(1)
                        crmax1_linear = poly_refined(naxis1)
                        cdelt1_linear = (crmax1_linear - crmin1_linear) / \
                                        (naxis1 - 1)
                        self.logger.info('linear crval1, cdelt1: %f %f',
                                         crmin1_linear, cdelt1_linear)
                        self.logger.info('fitted coefficients %s',
                                         poly_refined.coef)
                        self.logger.info('npoints_eff, residual_std: %d %f',
                                         npoints_eff, residual_std)
                        cr_linear = CrLinear(
                            crpix=1.0,
                            crval=crmin1_linear,
                            crmin=crmin1_linear,
                            crmax=crmax1_linear,
                            cdelt=cdelt1_linear
                        )
                        solution_wv = SolutionArcCalibration(
                            features=[],  # empty list!
                            coeff=poly_refined.coef,
                            residual_std=residual_std,
                            cr_linear=cr_linear
                        )
                        solution_wv.npoints_eff = npoints_eff  # add also this
                        new = FiberSolutionArcCalibration(fibid, solution_wv)
                        data_wlcalib.contents.append(new)
                        # store results for plotting
                        plot_tracenumber.append(fibid)
                        plot_npointseff.append(npoints_eff)
                        plot_residualstd.append(residual_std)
                        plot_crval1.append(crmin1_linear)
                        plot_cdelt1.append(cdelt1_linear)
                        plot_coeff.append(poly_refined.coef)
                    else:
                        self.logger.error('error in row %d, fibid %d',
                                          idx, fibid)
                        data_wlcalib.error_fitting.append(fibid)
                else:
                    self.logger.info('skipping row %d, fibid %d, not extracted',
                                     idx, fibid)
                    missing_fib += 1
                    data_wlcalib.missing_fibers.append(fibid)

            self.logger.info('Errors in fitting: %s', error_contador)
            self.logger.info('Missing fibers: %s', missing_fib)

            # save PDF file with plots in working directory
            if self.intermediate_results:
                from numina.array.display.matplotlib_qt import plt
                from matplotlib.backends.backend_pdf import PdfPages
                pdf = PdfPages('wavecal_iter2.pdf')
                for dumplot in zip(
                        [plot_npointseff, plot_residualstd, plot_crval1,
                         plot_cdelt1],
                        ['effective number of lines found',
                         'residual std ' + r'($\AA$)',
                         'linear CRVAL1 ' + r'($\AA$)',
                         'linear_CDELT1 ' + r'($\AA$/pixel)']):
                    ax = ximplotxy(plot_tracenumber, dumplot[0],
                                   xlabel='fiber number', ylabel=dumplot[1],
                                   linestyle='', marker='.', color='C0',
                                   show=False)
                    pdf.savefig()
                    plt.close()
                for ideg in range(poldeg_refined + 1):
                    dumplot = [coef[ideg] for coef in plot_coeff]
                    ax = ximplotxy(plot_tracenumber, dumplot,
                                   xlabel='fiber number',
                                   ylabel='coef[' + str(ideg) + ']',
                                   linestyle='', marker='.', color='C0',
                                   show=False)
                    pdf.savefig()
                    plt.close()
                pdf.close()
        else:
            data_wlcalib = None

        self.logger.info('End arc calibration')

        return initial_data_wlcalib, data_wlcalib, fwhm_hdulist

    def generate_fwhm_image(self, solutions):
        from scipy.spatial import cKDTree

        # Each 10 fibers. Comment this to iterate over all fibers instead.
        ##################################################################
        aux = []
        for value in solutions:
            if value.fibid % 10 == 0:
                aux.append(value)

        ##################################################################

        final = []
        for fibercalib in aux:
            for feature in fibercalib.solution.features:
                final.append([feature.xpos, feature.ypos, feature.fwhm])
        final = numpy.asarray(final)

        voronoi_points = numpy.array(final[:, [0, 1]])

        # Cartesian product of X, Y
        x = numpy.arange(2048 * 2)
        y = numpy.arange(2056 * 2)
        test_points = numpy.transpose(
            [numpy.tile(x, len(y)), numpy.repeat(y, len(x))])

        voronoi_kdtree = cKDTree(voronoi_points)

        test_point_dist, test_point_regions = voronoi_kdtree.query(test_points,
                                                                   k=1)
        final_image = test_point_regions.reshape((4112, 4096)).astype('float32')
        final_image[:, :] = final[final_image[:, :].astype('int16'), 2]
        return (final_image)

    def model_coeff_vs_fiber(self, data_wlcalib, poldeg,
                             times_sigma_reject=5,
                             debugplot=0):
        """Model polynomial coefficients vs. fiber number.

        For each polynomial coefficient, a smooth polynomial dependence
        with fiber number is also computed, rejecting information from
        fibers which coefficients depart from that smooth variation.
        """

        list_fibid = []
        list_coeffs = []
        for item in (data_wlcalib.contents):
            list_fibid.append(item.fibid)
            if len(item.solution.coeff) != poldeg + 1:
                raise ValueError('Unexpected number of polynomial '
                                 'coefficients')
            list_coeffs.append(item.solution.coeff)

        # determine bad fits from each independent polynomial coefficient
        poldeg_coeff_vs_fiber = 5
        reject_all = None  # avoid PyCharm warning
        fibid = numpy.array(list_fibid)
        for i in range(poldeg + 1):
            coeff = numpy.array([coeff[i] for coeff in list_coeffs])
            poly, yres, reject = polfit_residuals_with_sigma_rejection(
                x=fibid,
                y=coeff,
                deg=poldeg_coeff_vs_fiber,
                times_sigma_reject=times_sigma_reject,
            )
            if abs(debugplot) % 10 != 0:
                polfit_residuals(
                    x=fibid,
                    y=coeff,
                    deg=poldeg_coeff_vs_fiber,
                    reject=reject,
                    xlabel='fibid',
                    ylabel='coeff a_' + str(i),
                    title='Identifying bad fits',
                    debugplot=debugplot
                )
            if i == 0:
                reject_all = numpy.copy(reject)
                if abs(debugplot) >= 10:
                    print('coeff a_' + str(i) + ': nreject=', sum(reject_all))
            else:
                # add new bad fits
                reject_all = numpy.logical_or(reject_all, reject)
                if abs(debugplot) >= 10:
                    print('coeff a_' + str(i) + ': nreject=', sum(reject_all))

        # determine new fits excluding all fibers with bad fits
        list_poly_vs_fiber = []
        for i in range(poldeg + 1):
            coeff = numpy.array([coeff[i] for coeff in list_coeffs])
            poly, yres = polfit_residuals(
                x=fibid,
                y=coeff,
                deg=poldeg_coeff_vs_fiber,
                reject=reject_all,
                xlabel='fibid',
                ylabel='coeff a_' + str(i),
                title='Computing filtered fits',
                debugplot=debugplot
            )
            list_poly_vs_fiber.append(poly)

        if abs(debugplot) >= 10:
            print("list_poly_vs_fiber:\n", list_poly_vs_fiber)

        return list_poly_vs_fiber




import pytest
import astropy.io.fits as fits
import numpy

from ..multirss import generate_multi_rss

def generate_imgs(nimages, nfibers, nsamples):
    imgs = []
    for _ in range(nimages):
        data = numpy.empty((nfibers, nsamples))
        hdu1 = fits.PrimaryHDU(data)
        hdu2 = fits.ImageHDU(name='FIBERS')
        img = fits.HDUList([hdu1, hdu2])
        imgs.append(img)
    return imgs


def test_multirss():

    nimages = 10
    nfibers = 12
    nsamples = 300
    imgs = generate_imgs(nimages, nfibers, nsamples)

    result = generate_multi_rss(imgs)

    assert len(result) == nimages + 1

    primary = result[0]
    data = primary.data
    assert data.shape == (nimages * nfibers, nsamples)
    assert primary.header['MEG-NRSS'] == nimages
    # Check
    for idx, ext in enumerate(result[1:], 1):
        assert ext.header['EXTNAME'] == "FIBERS{}".format(idx)


def test_multirss_error():

    nimages = 0
    nfibers = 12
    nsamples = 300
    imgs = generate_imgs(nimages, nfibers, nsamples)

    with pytest.raises(ValueError):
        generate_multi_rss(imgs)



#
# Copyright 2011-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Fiber tracing Recipe."""

from __future__ import division, print_function

import logging
from datetime import datetime

import numpy
import numpy.polynomial.polynomial as nppol
from numina.array.peaks.peakdet import refine_peaks
from numina.array.trace.traces import trace
from numina.core import Product, Parameter
import matplotlib.pyplot as plt

import numina.core.qc as qc
from numina.array import combine
import numina.core.validator
from skimage.filters import threshold_otsu
from skimage.feature import peak_local_max
from scipy.ndimage.filters import minimum_filter
from numina.frame.utils import copy_img

from megaradrp.processing.aperture import ApertureExtractor
from megaradrp.processing.combine import basic_processing_with_combination
from megaradrp.products import TraceMap
from megaradrp.products.tracemap import GeometricTrace
from megaradrp.types import ProcessedImage, ProcessedRSS
from megaradrp.core.recipe import MegaraBaseRecipe
import megaradrp.requirements as reqs
import megaradrp.products
import megaradrp.processing.fibermatch as fibermatch
from megaradrp.instrument import vph_thr


class TraceMapRecipe(MegaraBaseRecipe):
    """Provides tracing information from continuum flat images.

    This recipe process a set of continuum flat images obtained in
    *Trace Map* mode and returns the tracing information required
    to perform fiber extraction in other recipes. The recipe also
    returns the result of processing the input images upto dark correction.

    See Also
    --------
    megaradrp.products.tracemap.TraceMap: description of TraceMap product
    megaradrp.recipes.calibration.modelmap.ModelMapRecipe: description of ModelMap recipe
    numina.array.trace.traces: tracing algorithm
    megaradrp.instrument.configs: instrument configuration

    Notes
    -----
    Images provided in `obresult` are trimmed and corrected from overscan,
    bad pixel mask (if `master_bpm` is not None), bias and dark current
    (if `master_dark` is not None).
    Images thus corrected are the stacked using the median.

    The result of the combination is saved as an intermediate result, named
    'reduced_image.fits'. This combined image is also returned in the field
    `reduced_image` of the recipe result and will be used for
    tracing the position of the fibers.

    The fibers are grouped in packs of different numbers of fibers. To match
    the traces in the image with the corresponding fibers is neccessary
    to know how fibers are packed and where the different groups of fibers
    appear in the detector. This information is provided by the fields
    'pseudoslit.boxes' and 'pseudoslit.boxes_positions' of the instrument
    configuration.

    Using the column reference provided by 'pseudoslit.boxes_positions', peaks
    are detected (using an average of 7 columns) and matched to the layout
    of fibers provided by 'pseudoslit.boxes_positions'. Fibers without a matching peak
    are counted and their ids stored in the final `master_traces` object.

    Once the peaks in the reference column are found, each one is traced
    until the border of the image is reached. The trace may be lost before
    reaching the border. In all cases, the beginning and the end of the trace
    are stored.

    The Y position of the trace is fitted to a polynomial
    of degree `polynomial_degree`. The coefficients of the polynomial are
    stored in the final `master_traces` object.

    """
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_bpm = reqs.MasterBPMRequirement()
    polynomial_degree = Parameter(5, 'Polynomial degree of trace fitting')
    relative_threshold = Parameter(0.3, 'Threshold for peak detection')
    debug_plot = Parameter(0, 'Save intermediate tracing plots')

    reduced_image = Product(ProcessedImage)
    reduced_rss = Product(ProcessedRSS)
    master_traces = Product(TraceMap)

    def run_qc(self, recipe_input, recipe_result):
        """Run quality control checks"""
        self.logger.info('start trace recipe QC')
        recipe_result.qc = qc.QC.GOOD
        recipe_result.master_traces.quality_control = qc.QC.GOOD
        self.logger.info('end trace recipe QC')
        return recipe_result

    @numina.core.validator.validate
    def run(self, rinput):
        """Execute the recipe.

        Parameters
        ----------
        rinput : TraceMapRecipe.RecipeInput

        Returns
        -------
        TraceMapRecipe.RecipeResult

        Raises
        ------
        ValidationError
              if the recipe input is invalid

        """
        self.logger.info('start trace spectra recipe')

        obresult = rinput.obresult
        current_vph = obresult.tags['vph']
        current_insmode = obresult.tags['insmode']
        obresult_meta = obresult.metadata_with(self.datamodel)

        debug_plot = rinput.debug_plot if self.intermediate_results else 0

        self.logger.info('start basic reduction')
        flow = self.init_filters(rinput, obresult.configuration)
        reduced = basic_processing_with_combination(rinput, flow, method=combine.median)
        self.logger.info('end basic reduction')

        self.save_intermediate_img(reduced, 'reduced_image.fits')

        insconf = obresult.configuration

        boxes = insconf.get('pseudoslit.boxes', **obresult.tags)

        box_borders0, cstart0 = self.obtain_boxes(insconf, obresult.tags)

        box_borders, cstart = self.refine_boxes_from_image(reduced, box_borders0, cstart0)

        self.logger.debug("original boxes: %s", box_borders0)
        self.logger.debug("refined boxes: %s", box_borders)

        if current_insmode in vph_thr and current_vph in vph_thr[current_insmode]:
            threshold = vph_thr[current_insmode][current_vph]
            self.logger.info('rel threshold for %s is %4.2f', current_vph, threshold)
        else:
            threshold = rinput.relative_threshold
            self.logger.info('rel threshold not defined for %s, using %4.2f', current_vph, threshold)

        final = megaradrp.products.TraceMap(instrument=obresult.instrument)
        fiberconf = self.datamodel.get_fiberconf(reduced)
        final.total_fibers = fiberconf.nfibers
        final.tags = obresult.tags
        final.boxes_positions = box_borders
        final.ref_column = cstart

        final.update_metadata(self)
        final.update_metadata_origin(obresult_meta)
        # Temperature in Celsius with 2 decimals
        final.tags['temp'] = round(obresult_meta['info'][0]['temp'] - 273.15, 2)

        contents, error_fitting = self.search_traces(
            reduced,
            boxes,
            box_borders,
            cstart=cstart,
            threshold=threshold,
            poldeg=rinput.polynomial_degree,
            debug_plot=debug_plot
        )

        final.contents = contents
        final.error_fitting = error_fitting

        # Perform extraction with own traces
        calibrator_aper = ApertureExtractor(final, self.datamodel)
        reduced_copy = copy_img(reduced)
        reduced_rss = calibrator_aper(reduced_copy)

        if self.intermediate_results:
            with open('ds9.reg', 'w') as ds9reg:
                final.to_ds9_reg(ds9reg, rawimage=False,
                                 numpix=100, fibid_at=2048)

            with open('ds9_raw.reg', 'w') as ds9reg:
                final.to_ds9_reg(ds9reg, rawimage=True,
                                 numpix=100, fibid_at=2048)

        self.logger.info('end trace spectra recipe')
        return self.create_result(reduced_image=reduced,
                                  reduced_rss = reduced_rss,
                                  master_traces=final)

    def obtain_boxes(self, insconf, tags):
        values = insconf.get('pseudoslit.boxes_positions', **tags)
        cstart = values['ref_column']
        box_borders = values['positions']
        return box_borders, cstart

    def obtain_boxes_from_image(self, reduced, expected, npeaks, cstart=2000):
        from numina.array.peaks.peakdet import find_peaks_indexes
        col = cstart
        data = reduced[0].data
        rr = data[:, col-1:col+1].mean(axis=1)
        # standarize
        rr -= numpy.median(rr)
        rr /= rr.max()
        rr *= -1

        cb = cosinebell(len(rr), 0.10)
        cbr = cb * rr
        plt.plot(cbr)
        plt.show()
        xv = numpy.fft.fftfreq(len(cbr))
        yv = numpy.fft.fft(cbr)
        plt.xlim([0.0, 0.5])
        plt.semilogy(xv, numpy.abs(yv.real))
        plt.show()

        cut = abs(xv) > 0.1
        yv[cut] = 0
        res = numpy.fft.ifft(yv)
        final = res.real
        plt.plot(final)
        #trend = detrend(final)
        #plt.plot(final - trend)
        plt.show()

        idx = find_peaks_indexes(final, window_width=3, threshold=0.3, fpeak=1)

        # We expect  differentnumber of pekas in LCB/MOS
        # order by intensity
        peak_flux = final[idx]
        # Number of peaks must be >=18
        npeaks = npeaks + 1
        fidx = numpy.argsort(peak_flux)[:-(npeaks+1):-1]
        nidx = idx[fidx]
        nidxs = numpy.sort(nidx)

        plt.plot(final)
        #plt.scatter(idx, [0.9 for m in idx])
        plt.scatter(nidx, [0.95 for m in nidx], c='r')
        plt.scatter(expected, [1.0 for m in expected])
        plt.show()

        plt.scatter(expected, nidxs - expected)
        plt.show()

        print("expected", expected)
        print("nidx", nidxs)

        return nidxs, col

    def refine_boxes_from_image(self, reduced, expected, cstart=2000, nsearch=20):
        """Refine boxes using a filtered Fourier image"""

        hs = 3
        # Cut freq in Fourier space
        cut_frec = 0.10
        # Cosine bell
        cos_cut = 0.10

        data = reduced[0].data
        rr = data[:, cstart-hs:cstart+hs].mean(axis=1)

        # standarize and Y flip
        rr -= numpy.median(rr)
        rr /= rr.max()
        rr *= -1

        cb = cosinebell(len(rr), cos_cut)
        cbr = cb * rr

        xv = numpy.fft.fftfreq(len(cbr))
        yv = numpy.fft.fft(cbr)

        # Filter freqs in Fourier space
        yv[abs(xv) > cut_frec] = 0

        res = numpy.fft.ifft(yv)
        final = res.real

        refined = expected[:]

        for ibox, box in enumerate(expected):
            iargmax = final[box - nsearch: box + nsearch +1].argmax()
            refined[ibox] = iargmax + box - nsearch

        return refined, cstart

    def search_traces(self, reduced, boxes, box_borders, cstart=2000,
                      threshold=0.3, poldeg=5, step=2, debug_plot=0):

        data = reduced[0].data

        hs = 3
        tol = 1.63

        self.logger.info('search for traces')

        self.logger.info('estimate background in ref column %i', cstart)
        background = estimate_background(data, center=cstart, hs=hs, boxref=box_borders)
        self.logger.info('background level is %f', background)

        self.logger.info('find peaks in reference column %i', cstart)

        central_peaks = init_traces(
            data,
            center=cstart,
            hs=hs,
            boxes=boxes,
            box_borders=box_borders,
            tol=tol,
            threshold=threshold,
            debug_plot=debug_plot
        )

        # The byteswapping is required by the cython module
        if data.dtype.byteorder != '=':
            self.logger.debug('byteswapping image')
            image2 = data.byteswap().newbyteorder()
        else:
            image2 = data

        maxdis = 2.0

        contents = []
        error_fitting = []
        self.logger.info('trace peaks from references')
        for dtrace in central_peaks:
            # FIXME, for traces, the background must be local
            # the background in the center is not always good
            local_trace_background = 300  # background

            self.logger.debug('trace fiber %d', dtrace.fibid)
            if dtrace.start:
                mm = trace(image2, x=cstart, y=dtrace.start[1], step=step,
                           hs=hs, background=local_trace_background, maxdis=maxdis)

                if debug_plot:
                    plt.plot(mm[:, 0], mm[:, 1], '.')
                    plt.savefig('trace-xy-%d.png' % dtrace.fibid)
                    plt.close()
                    plt.plot(mm[:, 0], mm[:, 2], '.')
                    plt.savefig('trace-xz-%d.png' % dtrace.fibid)
                    plt.close()
                if len(mm) < poldeg + 1:
                    self.logger.warning('in fibid %d, only %d points to fit pol of degree %d',
                                        dtrace.fibid, len(mm), poldeg)
                    pfit = numpy.array([])
                else:
                    pfit = nppol.polyfit(mm[:, 0], mm[:, 1], deg=poldeg)

                start = mm[0, 0]
                stop = mm[-1, 0]
            else:
                pfit = numpy.array([])
                start = cstart
                stop = cstart
                error_fitting.append(dtrace.fibid)

            self.logger.debug('trace start %d  stop %d', int(start), int(stop))

            this_trace = GeometricTrace(
                fibid=dtrace.fibid,
                boxid=dtrace.boxid,
                start=int(start),
                stop=int(stop),
                fitparms=pfit.tolist()
            )
            contents.append(this_trace)

        return contents, error_fitting


def estimate_background(image, center, hs, boxref):
    """Estimate background from values in boxes between fibers"""

    cut_region = slice(center-hs, center+hs)
    cut = image[boxref, cut_region]

    colcut = cut.mean(axis=1)

    return threshold_otsu(colcut)


# FIXME: need a better place for this
# Moved from megaradrp.trace


class FiberTraceInfo(object):
    def __init__(self, fibid, boxid):
        self.boxid = boxid
        self.fibid = fibid
        self.start = None


def init_traces(image, center, hs, boxes, box_borders, tol=1.5, threshold=0.37, debug_plot=0):

    _logger = logging.getLogger(__name__)

    cut_region = slice(center-hs, center+hs)
    cut = image[:, cut_region]
    colcut = cut.mean(axis=1)
    # trace local background with a min filter
    mincol = minimum_filter(colcut, size=7)

    _logger.debug('initial pairing fibers in column %d', center)

    fiber_traces = []
    total_peaks = 0
    lastid = 0
    counted_fibers = 0
    boxes_with_missing_fibers = []

    for boxid, box in enumerate(boxes):
        nfibers = box['nfibers']
        mfibers = box.get('missing', [])
        nfibers_max = nfibers - len(mfibers)
        sfibers = box.get('skipcount', [])
        _logger.debug('pseudoslit box: %s, id: %d', box['name'], boxid)
        _logger.debug('nfibers: %d, missing: %s',nfibers, mfibers)

        counted_fibers += nfibers
        b1 = int(box_borders[boxid])
        b2 = int(box_borders[boxid + 1])
        _logger.debug('box borders: %s %s', b1, b2)
        borders = [b1, b2]

        region = colcut[borders[0]:borders[1]+1]
        # Region for background computation
        # Remove the space of 1.5 fibers
        bb1 = b1 + 9
        bb2 = b2 - 9
        # a conservative background
        lowt = numpy.min(mincol[bb1:bb2+1])
        _logger.debug('conservative background in box is %f', lowt)

        # Find exactly the number of peaks expected
        _logger.debug('find %d peaks (max)', nfibers_max)
        ipeaks_int = peak_local_max(region, min_distance=3,
                                    threshold_abs=lowt,
                                    threshold_rel=threshold,
                                    num_peaks=nfibers_max,
                                    )[:, 0]

        npeaks = len(ipeaks_int)
        total_peaks += npeaks

        if npeaks == 0:
            # skip everything, go to next box
            boxes_with_missing_fibers.append((box['name'], nfibers))
            _logger.debug('no peaks detected, go to next box')
            continue

        # We always want the result sorted. The order changes in different versions
        # of scikit-image
        ipeaks_int.sort()
        ipeaks_float = refine_peaks(region, ipeaks_int, 3)[0]
        peaks_dist = numpy.diff(ipeaks_float)
        measured_scale = numpy.median(peaks_dist)
        _logger.debug('median distance between peaks is %s', measured_scale)

        peaks_y = numpy.ones((ipeaks_int.shape[0], 3))
        peaks_y[:, 0] = ipeaks_int + b1
        peaks_y[:, 1] = ipeaks_float + b1
        peaks_y[:, 2] = region[ipeaks_int]

        if debug_plot:
            plt.plot(region)
            plt.plot(ipeaks_int, region[ipeaks_int], 'r*')
            plt.savefig('central_cut_%d.png' % boxid)
            plt.close()

        startid = lastid + 1
        fiber_model = fibermatch.generate_box_model(nfibers,
                                                    start=startid,
                                                    missing_relids=mfibers,
                                                    skip_fibids=sfibers
                                                    )
        lastid = fiber_model[-1].fibid
        _logger.debug('fibids %s - %s', startid, lastid)

        matched_peaks = fibermatch.count_peaks(peaks_y[:, 1])
        nmatched = len(matched_peaks)
        missing = nfibers - nmatched

        if missing != 0:
            boxes_with_missing_fibers.append((box['name'], missing))

        _logger.debug('matched %s, missing: %s', nmatched, missing)
        pos_solutions = fibermatch.complete_solutions(fiber_model, matched_peaks,
                                                      borders, scale=measured_scale)

        for fibid, match in fibermatch.iter_best_solution(fiber_model,
                                               matched_peaks,
                                               pos_solutions):
            fti = FiberTraceInfo(fibid, boxid)
            if match is not None:
                fti.start = (center, peaks_y[match, 1], peaks_y[match, 2])
            fiber_traces.append(fti)

    _logger.debug('total found peaks: %d', total_peaks)
    _logger.debug('total found + recovered peaks: %d', counted_fibers)
    if boxes_with_missing_fibers:
        for m1, n2 in boxes_with_missing_fibers:
            _logger.debug('missing %d fibers in box %s', n2, m1)
    return fiber_traces


def cosinebell(n, fraction=0.10):
    """"Cosine bell mask"""
    mask = numpy.ones(n)
    nmasked = int(fraction*n)
    for i in range(nmasked):
        f = 0.5 * (1 - numpy.cos(numpy.pi * float(i) / float(nmasked)))
        mask[i] = f
        mask[n-i-1] = f
    return mask


#
# Copyright 2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import math

import numpy


def compute_pa_from_ipa(ipa, ins_angle=-163.256):
    """Recompute the PA from IPA """
    # get IPA keyword
    pa = -ins_angle + ipa
    return pa


def update_wcs_from_ipa(hdr, pa):
    """Recompute the WCS rotations from PA"""
    pa_rad = numpy.deg2rad(pa)
    cos_pa = math.cos(pa_rad)
    sin_pa = math.sin(pa_rad)

    # Update PC_ keywords
    hdr['PC1_1'] = cos_pa
    hdr['PC2_2'] = cos_pa
    hdr['PC1_2'] = sin_pa
    hdr['PC2_1'] = -sin_pa
    # CDELT1 must be negative
    hdr['CDELT1'] = -abs(hdr['CDELT1'])

    return hdr


#
# Copyright 2015-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import logging

import numpy
from numpy.lib.stride_tricks import as_strided as ast
from numina.simulation.detector import DetectorBase, VirtualDetector


_logger = logging.getLogger(__name__)


def binning(arr, br, bc):
    """Return a binned view if 'arr'"""
    nr, nc = arr.shape

    if nr % br != 0 or nc % bc != 0:
        raise ValueError("'bin' must be an integer multiple of size")

    bnr = nr // br
    bnc = nc // bc
    m = arr.dtype.itemsize
    newshape = bnr, bnc, br, bc
    newstrides = nc*br*m, bc*m, nc*m, m
    binned = ast(arr, shape=newshape, strides=newstrides)
    return binned


class ReadParams(object):
    """Readout parameters of each channel."""
    def __init__(self, gain=1.0, ron=2.0, bias=1000.0):
        self.gain = gain
        self.ron = ron
        self.bias = bias


class MegaraDetector(DetectorBase):
    """Simple MEGARA detector."""

    _binning = {'11': [1, 1], '21': [1, 2], '12': [2, 1], '22': [2, 2]}
    _direc = ['normal', 'mirror']

    def __init__(self, name, shape, oscan, pscan,
                 qe=1.0, qe_wl=None, dark=0.0,
                 readpars1=None, readpars2=None,
                 bins='11', direction='normal'):

        super(MegaraDetector, self).__init__(name, shape, qe, qe_wl, dark)

        self.oscan = oscan
        self.pscan = pscan

        self.readpars1 = readpars1 if not None else ReadParams()
        self.readpars2 = readpars2 if not None else ReadParams()

        self.bins = self._set_direction(direction)
        self._set_binning(bins)
        self.set_geometry()

    def configure(self, profile):
        if 'bins' in profile:
            self.set_binning(profile['bins'])

    def _set_binning(self, bins):

        if bins not in self._binning:
            raise ValueError("%s must be one if '11', '12', '21, '22'" % bins)

        self.bins = bins
        self.blocks = self._binning[self.bins]

    def set_binning(self, bins):

        self._set_binning(bins)
        self.set_geometry()

    def _set_direction(self, direction):

        if direction not in self._direc:
            raise ValueError("%s must be either 'normal' or 'mirror'" % direction)

        if direction == 'normal':
            directfun = lambda x: x
        else:
            directfun = numpy.fliplr

        self.directfun = directfun

    def set_direction(self, direction):

        self._set_direction(direction)
        self.set_geometry()

    def set_geometry(self):
        self.fshape, (base1, base2), geom1, geom2 = self.init_regions(self.dshape,
                                                                      self.oscan, self.pscan,
                                                                      self.blocks)

        self.virt1 = VirtualDetector(base1, geom1, self.directfun, self.readpars1)
        self.virt2 = VirtualDetector(base2, geom2, self.directfun, self.readpars2)

    def pre_readout(self, elec_pre):
        # FIXME: there is a bug in numpy here
        # this crashes if you don't make a copy()
        elec_pre_inv = numpy.flipud(elec_pre).copy()
        elec_v = binning(elec_pre_inv, self.blocks[0], self.blocks[1])
        elec_p = elec_v.reshape(elec_v.shape[0], elec_v.shape[1], -1)
        elec_f = elec_p.sum(axis=-1)
        return elec_f

    def base_readout(self, elec_f):
        # Output image
        final = numpy.zeros(self.fshape)
        self.virt1.readout_in_buffer(elec_f, final)
        self.virt2.readout_in_buffer(elec_f, final)
        return final

    def saturate(self, x):
        """Compute non-linearity and saturation."""
        a = -1.2 / 45000.0
        # b = 1.0
        # c = -7000.0
        # det = math.sqrt(b**2-4*a*c)
        # u1 = (-b+det) / (2*a)

        f1 = lambda x:x

        def f2(xx):
            return xx + a* (xx - 45000)**2

        f3 = lambda x: 52000

        p1 = x < 45000
        p3 = x >= 54312  # 45000 + u1
        p2 = numpy.logical_and(~p1, ~p3)

        return numpy.piecewise(x, [p1, p2, p3], [f1, f2, f3])

    # @classmethod
    def init_regions(cls, detshape, oscan, pscan, bng):
        """Create a image with overscan for testing."""

        base_rows = detshape[0] // 2
        base_cols = detshape[1] // 2
        base_oscan = oscan
        base_pscan = pscan

        nr = base_rows // bng[0]
        nc = base_cols // bng[1]

        nr2 = 2 * nr
        nc2 = 2 * nc

        oscan1 = base_oscan // bng[0]
        oscan2 = oscan1 * 2

        psc1 = base_pscan // bng[0]
        psc2 = 2 * psc1
        fshape = (nr2 + oscan2, nc2 + psc2)

        # Row block 1
        rb1 = slice(0, nr)
        rb1m = slice(nr, nr + oscan1)

        # Row block 2
        rb2 = slice(nr + oscan2, nr2 + oscan2)
        rb2m = slice(nr + oscan1, nr + oscan2)

        # Col block
        cb = slice(psc1, nc2 + psc1)
        # Col block left
        cbl = slice(0, psc1)
        # Col block right
        cbr = slice(nc2 + psc1, nc2 + psc2)
        # Col block full
        cbf = slice(0, nc2 + psc2)

        # Mode normal
        trim1 = (rb1, cb)
        # prescan
        pcol1 = (rb1, cbl)
        # overscan
        ocol1 = (rb1, cbr)
        orow1 = (rb1m, cbf)

        trim2 = (rb2, cb)
        # prescan
        pcol2 = (rb2, cbr)
        # overscan
        ocol2 = (rb2, cbl)
        orow2 = (rb2m, cbf)

        base1 = (slice(0, nr), slice(0, nc2))
        base2 = (slice(nr, nr2), slice(0, nc2))

        geom1 = trim1, pcol1, ocol1, orow1
        geom2 = trim2, pcol2, ocol2, orow2

        return (fshape, (base1, base2), geom1, geom2)

    def init_config_info(self):
        info = super(MegaraDetector, self).init_config_info()
        info['exposed'] = self._time_last
        info['vbin'] = int(self.bins[-1])
        info['hbin'] = int(self.bins[0])
        return info


class MegaraDetectorSat(MegaraDetector):

    def saturate(self, x):
        y = super(MegaraDetectorSat, self).saturate(x)

        # Some pixels have a special nonlineary given by
        # y[100,3000] = self.esp_nonlinearity(x[100,3000])
        # y[3000,3000] = self.esp_nonlinearity(x[3000,3000])
        return y

    def esp_nonlinearity(self, x):
        sat = 12000.0
        return sat * (1-sat / (sat + x))




#
# Copyright 2016-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Validators for Observing modes"""

import sys

import six

from numina.exceptions import ValidationError


def validate_focus(obresult):
    """Validate FOCUS_SPECTROGRAPH"""
    image_groups = {}
    for idx, frame in enumerate(obresult.frames):
        with frame.open() as img:
            try:
                focus_val = img[0].header['FOCUS']
                # FIXME: This should be done using an scheme for MEGARA
                if not isinstance(focus_val, (int, float)):
                    raise ValidationError("FOCUS must be integer, not {}".format(type(focus_val)))
                if focus_val not in image_groups:
                    image_groups[focus_val] = []
                image_groups[focus_val].append(frame)
            except Exception:
                _type, exc, tb = sys.exc_info()
                six.reraise(ValidationError, exc, tb)

    if len(image_groups) < 2:
        raise ValidationError('We have only {} different focus in OB'.format(len(image_groups)))

    return True


#
# Copyright 2016-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

from numina.simulation.wheel import Wheel


class MegaraCalibrationUnit(Wheel):
    pass




#
# Copyright 2015-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import numpy as np

from megaradrp.simulation.detector import binning

def test_binning():
    nr = 6
    nc = 8
    bc = 2
    br = 2
    bnr = nr / br
    bnc = nc / bc
    arr = np.arange(nr*nc, dtype='int64').reshape(nr, nc)

    barr = binning(arr, bc, br)

    assert barr.shape == (bnr, bnc, br, bc)

    assert np.all(barr[0,0] == np.array([[0, 1], [8,9]]))

    carr = barr.reshape(barr.shape[0], barr.shape[1], -1)

    res = np.array([[ 18,  26,  34,  42], [ 82,  90,  98, 106], [146, 154, 162, 170]])

    assert np.all(carr.sum(axis=-1) == res)


if __name__ == "__main__":
    test_binning()


#
# Copyright 2015-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

from __future__ import print_function

from astropy.io import fits
import numpy as np

# row / column
_binning = {'11': [1, 1], '21': [1, 2], '12': [2, 1], '22': [2, 2]}
_direc = ['normal', 'mirror']


def create(image, direction='normal', bins='11'):
    """Create a image with overscan for testing."""

    if direction not in _direc:
        raise ValueError("%s must be either 'normal' or 'mirror'" % direction)

    if direction == 'normal':
        direcfun = lambda x: x
    else:
        direcfun = np.fliplr

    if bins not in _binning:
        raise ValueError("%s must be one if '11', '12', '21, '22'" % bins)

    bng = _binning[bins]

    nr = 2056 / bng[0]
    nc = 2048 / bng[1]

    nr2 = 2 * nr
    nc2 = 2 * nc

    oscan1 = 50 / bng[0]
    oscan2 = oscan1 * 2

    psc1 = 50 / bng[0]
    psc2 = 2 * psc1

    fshape = (nr2 + oscan2, nc2 + psc2)

    # Row block 1
    rb1 = slice(0, nr)
    rb1m = slice(nr, nr + oscan1)
    # Row block 2
    rb2 = slice(nr + oscan2, nr2 + oscan2)
    rb2m = slice(nr + oscan1, nr + oscan2)
    # Col block
    cb = slice(psc1, nc2 + psc1)
    # Col block left
    cbl = slice(0, psc1)
    # Col block right
    cbr = slice(nc2 + psc1, nc2 + psc2)

    # Mode normal
    trim1 = (rb1, cb)
    pcol1 = (rb1, cbl)
    ocol1 = (rb1, cbr)
    orow1 = (rb1m, cb)
    print(trim1, ocol1, orow1, pcol1)

    trim2 = (rb2, cb)
    pcol2 = (rb2, cbr)
    ocol2 = (rb2, cbl)
    orow2 = (rb2m, cb)
    print(trim2, ocol2, orow2, pcol2)

    finaldata = np.zeros(fshape, dtype='float32')

    finaldata[trim1] = direcfun(np.atleast_2d(np.arange(0, nc2)))
    finaldata[trim2] = direcfun(np.atleast_2d(np.arange(0, nc2)))

    finaldata[orow1] = 3
    finaldata[orow2] = 4

    finaldata[pcol1] = 5
    finaldata[pcol2] = 6

    finaldata[ocol1] = 7
    finaldata[ocol2] = 8

    hdu = fits.PrimaryHDU(data=finaldata)
    hdu.writeto(image, overwrite=True)



#
# Copyright 2011-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Products of the Megara Pipeline: Wavelength  Calibration"""


from numina.array.wavecalib.arccalibration import SolutionArcCalibration
import numpy.polynomial.polynomial as nppol

from .structured import BaseStructuredCalibration


class FiberSolutionArcCalibration(object):
    def __init__(self, fibid, solution):
        self.fibid = fibid
        self.solution = solution

    def __getstate__(self):
        return {'fibid': self.fibid,
                'solution': self.solution.__getstate__()
        }

    def __setstate__(self, state):
        self.fibid = state['fibid']
        new = SolutionArcCalibration.__new__(SolutionArcCalibration)
        new.__setstate__(state['solution'])
        self.solution = new


class WavelengthCalibration(BaseStructuredCalibration):
    """Wavelength Calibration Product
    """
    def __init__(self, instrument='unknown'):
        super(WavelengthCalibration, self).__init__(instrument)
        self.contents = []
        self.global_offset = nppol.Polynomial([0.0])

    def tag_names(self):
        return ['insmode', 'vph']

    def __getstate__(self):
        st = super(WavelengthCalibration, self).__getstate__()

        st['contents'] = [val.__getstate__() for val in self.contents]
        st['global_offset'] = self.global_offset.coef
        return st

    def __setstate__(self, state):
        super(WavelengthCalibration, self).__setstate__(state)

        self.global_offset = nppol.Polynomial(state.get('global_offset', [0.0]))

        self.contents = []
        # Handle dictionary
        if isinstance(state['contents'], dict):
            values = state['contents'].values()
        else:
            values = state['contents']

        for val in values:
            new = FiberSolutionArcCalibration.__new__(FiberSolutionArcCalibration)
            new.__setstate__(val)
            self.contents.append(new)



#
# Copyright 2011-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

from numina.tests.plugins import *




import pytest

import megaradrp.requirements as reqs
import numina.core
import numina.dal
from numina.exceptions import NoResultFound


class IoInstance(object):
    _numina_desc_val = {}


class LocalDal(object):
    def __init__(self):
        self.filename = "some_file.txt"

    def search_product(self, name, tipo, obsres, options=None):
        if obsres.id == 1:
            return numina.dal.StoredProduct(1, self.filename, tags={})
        if obsres.id == 2:
            return numina.dal.StoredProduct(1, None, tags={})
        elif obsres.id == 0:
            raise NoResultFound('Nothing')
        else:
            sp = None
        return sp


@pytest.mark.parametrize('obsid', [0, 1, 2])
def test_extinction_none(obsid):

    name = "reference_extinction"
    reference_extinction = reqs.ReferenceExtinction()
    reference_extinction.dest = name

    dal = LocalDal()

    obs = numina.core.ObservationResult()
    obs.id = obsid
    obs.requirements = {name: None}
    value = reference_extinction.query(dal, obs)
    instance = IoInstance()

    reference_extinction.__set__(instance, value)
    cval = instance._numina_desc_val[name]
    assert cval is None


def _test_extinction_filename():

    name = "reference_extinction"
    reference_extinction = reqs.ReferenceExtinction()
    reference_extinction.dest = name

    dal = LocalDal()

    obs = numina.core.ObservationResult()

    import tempfile, os, numpy

    arr = numpy.array([1.0,2.0,3.0,4.0])
    fd, path = tempfile.mkstemp()
    try:
        #obs.requirements = {name: path}
        numpy.savetxt(path, arr)

        value = reference_extinction.query(dal, obs)
        instance = IoInstance()

        reference_extinction.__set__(instance, value)
        cval = instance._numina_desc_val[name]
        assert numpy.allclose(cval, arr)
        assert False
    finally:
        os.remove(path)


#
# Copyright 2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Products of the Megara Pipeline"""

import numpy
import numpy.polynomial.polynomial as nppol

from numina.util.convertfunc import json_serial_function, convert_function

from megaradrp.products.structured import BaseStructuredCalibration
from .traces import to_ds9_reg as to_ds9_reg_function


class GeometricModel(object):
    def __init__(self, fibid, boxid, start, stop, model):
        self.fibid = fibid
        self.boxid = boxid
        self.start = start
        self.stop = stop
        self.model = model

    @property
    def valid(self):
        if self.model:
            return True
        else:
            return False

    def __getstate__(self):
        state = self.__dict__.copy()
        # del state['model']

        params = state['model']['params']
        newparams = {}
        for key, val in params.items():
            serial = json_serial_function(val)
            newparams[key] = serial

        state['model']['params'] = newparams
        return state

    def __setstate__(self, state):
        self.__dict__ = state
        self._set_model(state['model'])

    def _set_model(self, model):
        if model:
            params = {}
            for key, val in model['params'].items():
                params[key] =  convert_function(val)
            model['params'] = params

    @property
    def polynomial(self):
        # FIXME: this is a workaround
        return self.model['params']['mean']


class ModelMap(BaseStructuredCalibration):
    def __init__(self, instrument='unknown'):
        super(ModelMap, self).__init__(instrument)
        self.contents = []
        self.boxes_positions = []
        self.global_offset = nppol.Polynomial([0.0])
        self.ref_column = 2000
        self._wcols = None

    def __getstate__(self):
        st = super(ModelMap, self).__getstate__()
        st['contents'] = [t.__getstate__() for t in self.contents]
        st['boxes_positions'] = self.boxes_positions
        st['global_offset'] = self.global_offset.coef
        st['ref_column'] = self.ref_column
        return st

    def __setstate__(self, state):
        super(ModelMap, self).__setstate__(state)
        # self.contents = [GeometricModel(**trace) for trace in state['contents']]
        self.contents = []
        for trace in state['contents']:
            m = GeometricModel.__new__(GeometricModel)
            m.__setstate__(trace)
            self.contents.append(m)
        self.boxes_positions = state.get('boxes_positions', [])
        self.global_offset = nppol.Polynomial(state.get('global_offset', [0.0]))
        self.ref_column = state.get('ref_column', 2000)
        self._wcols = None

    def calculate_matrices(self, shape, processes=0):
        if self._wcols is None:
            self._wcols = calc_matrix_cols(self, shape, processes)

    def aper_extract(self, img, processes=0):
        if self._wcols is None:
            self._wcols = calc_matrix_cols(self, img.shape, processes)
        return aper_extract(self, self._wcols, img)

    def to_ds9_reg(self, ds9reg, rawimage=False, numpix=100, fibid_at=0):
        """Transform fiber traces to ds9-region format.

        Parameters
        ----------
        ds9reg : BinaryIO
            Handle to output file name in ds9-region format.
        rawimage : bool
            If True the traces must be generated to be overplotted on
            raw FITS images.
        numpix : int
            Number of abscissae per fiber trace.
        fibid_at : int
            Abscissae where the fibid is shown (default=0 -> not shown).

        """
        return to_ds9_reg_function(self, ds9reg,
                                   rawimage=rawimage,
                                   numpix=numpix,
                                   fibid_at=fibid_at
                                   )



# BUILD MATRICES
def calc_matrix(g_mean, g_std, valid, wshape, clip=1.0e-6, extra=10):
    from scipy.sparse import lil_matrix
    from numina.modeling.gaussbox import gauss_box_model
    # calc w
    # this is valid for 1 column (there are more broadcasts for N)

    begpix = numpy.ceil(g_mean - 0.5).astype('int')

    steps = numpy.arange(-extra, extra)
    # ref is a bidimensional matrix +-10 pixels around the trace
    ref = begpix + steps[:, numpy.newaxis]

    rr = gauss_box_model(ref, mean=g_mean, stddev=g_std)
    rrb = begpix - extra

    # This was sending warnings. Is there a NaN somewhere?
    with numpy.errstate(invalid='ignore'):
        rr[rr < clip] = 0.0

    # Calc Ws matrix
    block, valid_nfib = rr.shape

    w_init = lil_matrix(wshape)
    for fibid in valid:
        idx = fibid - 1
        w_init[rrb[idx]:rrb[idx] + block, idx] = rr[:, idx, numpy.newaxis]

    wcol = w_init.tocsr()
    return wcol


def calc_matrix_alt(g_mean, g_std, col, valid, wshape, clip=1.0e-6, extra=10):

    # For parallel processing
    wcol = calc_matrix(g_mean, g_std, valid, wshape, clip=clip, extra=extra)

    return col, wcol


def calc_matrix_cols(model_map, datashape, processes=0):

    dnrow, dncol = datashape
    nfibs = model_map.total_fibers

    shape = (nfibs, dncol)
    wshape = (dnrow, nfibs)

    array_mean = numpy.zeros(shape)
    array_std = numpy.zeros(shape)

    xcol = numpy.arange(dncol)
    for pesos in model_map.contents:
        if pesos.valid:
            row = pesos.fibid - 1
            params = pesos.model['params']
            array_mean[row] = params['mean'](xcol)
            array_std[row] = params['stddev'](xcol)

    # shift array_mean according to global_offset
    mask = numpy.zeros((model_map.total_fibers,), dtype='bool')
    valid_r = [(f.fibid - 1) for f in model_map.contents if f.valid]
    mask[valid_r] = True

    mean_at_ref = array_mean[mask, model_map.ref_column]
    offset = model_map.global_offset(mean_at_ref)
    array_mean[mask, :] = array_mean[mask, :] + offset[:, numpy.newaxis]

    wcols = {}
    valid = [f.fibid for f in model_map.contents if f.valid]
    if processes < 2:
        for col in xcol:
            result = calc_matrix(array_mean[:, col], array_std[:, col],
                                 valid, wshape, clip=1e-6, extra=10)
            wcols[col] = result
    else:
        import multiprocessing as mp
        pool = mp.Pool(processes=processes)

        results = [pool.apply_async(
            calc_matrix_alt,
            args=(array_mean[:, col], array_std[:, col], col, valid, wshape),
            kwds={'clip': 1e-6, 'extra': 10}
        ) for col in xcol]

        for p in results:
            col, wcol = p.get()
            wcols[col] = wcol

    return wcols


def aper_extract(model_map, wcols, img):
    from scipy.sparse.linalg import lsqr

    n0 = model_map.total_fibers
    n1 = img.shape[1]
    rss = numpy.zeros((n0, n1))
    for key, val in wcols.items():
        yl = img[:, key]
        res = lsqr(val, yl)
        rss[:, key] = res[0]

    return rss


#
# Copyright 2015-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Typical requirements of recipes"""

from numina.core import Requirement
from numina.types.multitype import MultiType

import megaradrp.types
import megaradrp.products
import megaradrp.products.modelmap


class MasterBiasRequirement(Requirement):
    def __init__(self, optional=False):
        super(MasterBiasRequirement,
              self).__init__(megaradrp.types.MasterBias,
                             'Master BIAS image',
                             optional=optional
                             )


class MasterBPMRequirement(Requirement):
    def __init__(self, optional=True):
        super(MasterBPMRequirement,
              self).__init__(megaradrp.types.MasterBPM,
                             'Master Bad Pixel Mask',
                             optional=optional
                             )


class MasterDarkRequirement(Requirement):
    def __init__(self, optional=True):
        super(MasterDarkRequirement,
              self).__init__(megaradrp.types.MasterDark, 'Master DARK image',
                             optional=optional)


class MasterFiberFlatRequirement(Requirement):
    def __init__(self):
        super(MasterFiberFlatRequirement,
              self).__init__(megaradrp.types.MasterFiberFlat,
                             'Master fiber flat calibration'
                             )


class MasterSlitFlatRequirement(Requirement):
    def __init__(self, optional=True):
        super(MasterSlitFlatRequirement,
              self).__init__(megaradrp.types.MasterSlitFlat,
                             'Master slit flat calibration',
                             optional=optional
                             )


class MasterTwilightRequirement(Requirement):
    def __init__(self, optional=True):
        super(MasterTwilightRequirement,
              self).__init__(megaradrp.types.MasterTwilightFlat,
                             'Master twlight flat calibration',
                             optional=optional
                             )


class MasterTraceMapRequirement(Requirement):
    def __init__(self):
        super(MasterTraceMapRequirement,
              self).__init__(megaradrp.products.TraceMap, 'Trace information of the Apertures')


class MasterAperturesRequirement(Requirement):
    def __init__(self):
        super(MasterAperturesRequirement, self).__init__(MultiType(
            megaradrp.products.modelmap.ModelMap,
            megaradrp.products.TraceMap), 'Apertures information for extraction')


class WavelengthCalibrationRequirement(Requirement):
    def __init__(self):
        super(WavelengthCalibrationRequirement,
              self).__init__(megaradrp.products.WavelengthCalibration, 'Wavelength calibration table')


class LinesCatalogRequirement(Requirement):
    def __init__(self):
        super(LinesCatalogRequirement, self).__init__(megaradrp.types.MegaraLinesCatalog, 'Catalog of lines')


class SensitivityRequirement(Requirement):
    def __init__(self, optional=True):
        super(SensitivityRequirement,
              self).__init__(megaradrp.types.MasterSensitivity,
                             'Master sensitivity for flux calibration',
                             optional=optional
                             )


class ReferenceExtinction(Requirement):
    def __init__(self, optional=True):
        super(ReferenceExtinction,
              self).__init__(megaradrp.types.ReferenceExtinctionTable,
                             "Reference extinction",
                             optional=optional
                             )


#
# Copyright 2011-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Bad Pixel Mask (BPM) recipe"""

import uuid
import datetime

import numina.exceptions
import astropy.io.fits as fits
from numina.array.cosmetics import ccdmask
from numina.core import Product
from numina.array import combine

from megaradrp.processing.combine import basic_processing_with_combination_frames
from megaradrp.core.recipe import MegaraBaseRecipe
from megaradrp.types import  MasterBPM
import megaradrp.requirements as reqs


class BadPixelsMaskRecipe(MegaraBaseRecipe):
    """Process defocussed FIBER_FLAT images and create MASTER_BPM product.

    This recipe process a set of defocused continuum flat images obtained in
    **Bad-pixels mask** mode and returns the master bad-pixels mask product.

    See Also
    --------
    numina.array.cosmetics.ccdmask: algorithm to select bad-pixels
    megaradrp.types.MasterBPM: description of MasterBPM product

    Notes
    -----
    Images provided in `obresult` are trimmed and corrected from overscan,
    bias and dark current (if `master_dark` is not None). The first
    half of the images are the stacked using the median and saved as
    intermediate result 'reduced_image_1.fits'. The second half is also
    combined and saved as intermediate result 'reduced_image_2.fits'

    These two images are passed to the `ccdmask` function, that selects
    bad-pixels by finding outliers in the ratio of the two images.

    The mask is returned in the field `master_bpm` of the recipe result.
    """

    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()

    master_bpm = Product(MasterBPM)

    def run(self, rinput):
        self.logger.info('start BPM recipe')
        N = len(rinput.obresult.frames)

        self.logger.debug('we have %d images', N)
        half = N // 2
        flow = self.init_filters(rinput, rinput.obresult.configuration)
        self.logger.debug('we have %d images', N)
        reduced1 = basic_processing_with_combination_frames(
            rinput.obresult.frames[:half],
            flow,
            method=combine.median
        )

        self.save_intermediate_img(reduced1, 'reduced_image_1.fits')

        reduced2 = basic_processing_with_combination_frames(
            rinput.obresult.frames[half:],
            flow,
            method=combine.median
        )

        self.save_intermediate_img(reduced2, 'reduced_image_2.fits')

        ratio, mask, sigma = ccdmask(reduced1[0].data, reduced2[0].data, mode='full')

        hdu = fits.PrimaryHDU(mask, header=reduced1[0].header)
        hdu.header['UUID'] = str(uuid.uuid1())

        self.set_base_headers(hdu.header)

        hdu.header['history'] = 'BPM creation time {}'.format(
            datetime.datetime.utcnow().isoformat()
        )

        for frame in rinput.obresult.frames:
            hdu.header['history'] = "With image {}".format(
                self.datamodel.get_imgid(frame.open())
            )

        reduced = fits.HDUList([hdu])
        self.logger.info('end BPM recipe')
        return self.create_result(master_bpm=reduced)

    def validate_input(self, recipe_input):
        """Validate input of the recipe.

        The number of frames in `recipe_input.obresult` must be even.

        Raises
        ------
        numina.exceptions.ValidationError
            If the number of frames in `obresult` is odd

        """

        obresult = recipe_input.obresult
        # Check that the number of frames is even
        nimages = len(obresult.frames)
        if nimages % 2 != 0:
            msg = 'expected even number of frames, received {} instead'.format(nimages)
            raise numina.exceptions.ValidationError(msg)
        # Continue with additional checks
        return super(BadPixelsMaskRecipe, self).validate_input(recipe_input)

    def set_base_headers(self, hdr):
        """Set metadata in FITS headers."""
        hdr = super(BadPixelsMaskRecipe, self).set_base_headers(hdr)
        hdr['NUMTYPE'] = ('MasterBPM', 'Product type')
        return hdr


#
# Copyright 2015-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import numina.core

import astropy.io.fits as fits

from megaradrp.taggers import tagger_empty, tagger_vph
from megaradrp.taggers import tagger_base_image


def test_tagger_empty():

    img1 = fits.PrimaryHDU(data=[1, 2, 3], header=fits.Header())
    img1.header['FILTER'] = 'FILTER-A'
    img1.header['READM'] = 'MOD1'
    frame1 = numina.core.DataFrame(frame=fits.HDUList(img1))

    ob = numina.core.ObservationResult()
    ob.frames = [frame1]

    tags = tagger_empty(ob)

    assert tags == {}


def test_tagger_empty_no_images():

    ob = numina.core.ObservationResult()

    tags = tagger_empty(ob)

    assert tags == {}


def test_tagger_vph():

    img1 = fits.PrimaryHDU(data=[1, 2, 3], header=fits.Header())
    img1.header['VPH'] = 'VPH405_LR'
    img1.header['READM'] = 'MOD1'
    frame1 = numina.core.DataFrame(frame=fits.HDUList(img1))

    img2 = fits.PrimaryHDU(data=[1, 2, 3], header=fits.Header())
    img2.header['VPH'] = 'VPH405_LR'
    img2.header['READM'] = 'MOD1'
    frame2 = numina.core.DataFrame(frame=fits.HDUList(img2))

    ob = numina.core.ObservationResult()
    ob.frames = [frame1, frame2]

    tags = tagger_vph(ob)

    assert tags == {'vph': 'VPH405_LR'}


def test_tagger_vph_no_images():

    ob = numina.core.ObservationResult()

    tags = tagger_vph(ob)

    assert tags == {}


def test_tagger_base_image():

    img1 = fits.PrimaryHDU(data=[1, 2, 3], header=fits.Header())
    img1.header['VPH'] = 'VPH405_LR'
    img1.header['INSMODE'] = 'MOS'
    img1.header['READM'] = 'MOD1'
    frame1 = numina.core.DataFrame(frame=fits.HDUList(img1))

    img2 = fits.PrimaryHDU(data=[1, 2, 3], header=fits.Header())
    img2.header['VPH'] = 'VPH405_LR'
    img2.header['INSMODE'] = 'MOS'
    img2.header['READM'] = 'MOD1'
    frame2 = numina.core.DataFrame(frame=fits.HDUList(img2))

    ob = numina.core.ObservationResult()
    ob.frames = [frame1, frame2]

    tags = tagger_base_image(ob)

    assert tags == {'vph': 'VPH405_LR', 'insmode': 'MOS'}


def test_tagger_base_image_no_images():

    ob = numina.core.ObservationResult()

    tags = tagger_base_image(ob)

    assert tags == {}

if __name__ == "__main__":
    test_tagger_vph()




from __future__ import print_function

from numina.core.pipelineload import DefaultLoader, build_instrument_config


class Loader(DefaultLoader):
    """Instrument configuration loader for MEGARA"""
    def __init__(self):
        super(Loader, self).__init__("megaradrp.instrument.configs")


if __name__ == '__main__':

    key = "4fd05b24-2ed9-457b-b563-a3c618bb1d4c"
    mm = build_instrument_config(key, loader=Loader())

    print('detector.scan', mm.get('detector.scan'))
    print('pseudoslit.boxes', mm.get('pseudoslit.boxes', insmode='LCB'))
    print('pseudoslit.boxes_positions', mm.get('pseudoslit.boxes_positions', insmode='LCB', vph='LR-I'))



#
# Copyright 2011-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Products of the Megara Pipeline"""

import numpy
import numpy.polynomial.polynomial as nppol

from .structured import BaseStructuredCalibration
from .traces import to_ds9_reg as to_ds9_reg_function


class GeometricTrace(object):
    def __init__(self, fibid, boxid, start, stop, fitparms=None):
        self.fibid = fibid
        self.boxid = boxid
        self.start = start
        self.stop = stop
        self.fitparms = fitparms if fitparms is not None else []
        self.polynomial = None
        # Update polynomial
        self._set_polynomial(fitparms)

    @property
    def valid(self):
        if self.fitparms:
            return True
        else:
            return False

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['polynomial']
        return state

    def __setstate__(self, state):
        self.__dict__ = state
        self._set_polynomial(state['fitparms'])

    def _set_polynomial(self, fitparms):
        if fitparms:
            self.polynomial = nppol.Polynomial(self.fitparms)
        else:
            self.polynomial = nppol.Polynomial([0.0])


class TraceMap(BaseStructuredCalibration):
    """Trace map calibration product"""
    def __init__(self, instrument='unknown'):
        super(TraceMap, self).__init__(instrument)
        self.contents = []
        self.boxes_positions = []
        self.global_offset = nppol.Polynomial([0.0])
        self.ref_column = 2000
        #
    def tag_names(self):
        return ['insmode', 'vph']

    def __getstate__(self):
        st = super(TraceMap, self).__getstate__()
        st['contents'] = [t.__getstate__() for t in self.contents]
        st['boxes_positions'] = self.boxes_positions
        st['global_offset'] = self.global_offset.coef
        st['ref_column'] = self.ref_column
        return st

    def __setstate__(self, state):
        super(TraceMap, self).__setstate__(state)
        self.contents = [GeometricTrace(**trace) for trace in state['contents']]
        self.boxes_positions = state.get('boxes_positions', [])
        self.global_offset = nppol.Polynomial(state.get('global_offset', [0.0]))
        self.ref_column = state.get('ref_column', 2000)
        return self

    def to_ds9_reg(self, ds9reg, rawimage=False, numpix=100, fibid_at=0):
        """Transform fiber traces to ds9-region format.

        Parameters
        ----------
        ds9reg : BinaryIO
            Handle to output file name in ds9-region format.
        rawimage : bool
            If True the traces must be generated to be overplotted on
            raw FITS images.
        numpix : int
            Number of abscissae per fiber trace.
        fibid_at : int
            Abscissae where the fibid is shown (default=0 -> not shown).

        """
        return to_ds9_reg_function(self, ds9reg,
                                   rawimage=rawimage,
                                   numpix=numpix,
                                   fibid_at=fibid_at
                                   )



#
# Copyright 2016-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#


from numina.simulation.wheel import Carrousel


class PseudoSlit(object):
    def __init__(self, name, insmode):

        super(PseudoSlit, self).__init__()
        self.name = name
        self.insmode = insmode
        self.fiberset = None
        self.positions = []

    def connect_fibers(self, fiberset, positions):
        self.fiberset = fiberset
        self.positions = positions

    def y_pos(self, fibsid):
        result = []
        for fibid in fibsid:
            pos = self.positions[fibid-1]
            result.append(pos)
        return result

    def config_info(self):
        return {'name': self.name,
                'insmode': self.insmode,
                'nfibers': len(self.fiberset.fibers),
                }


class PseudoSlitSelector(Carrousel):
    pass


#
# Copyright 2017-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import logging
import datetime


import astropy.wcs

import numpy
from numina.processing import Corrector


_logger = logging.getLogger(__name__)


PIXLIM_KEYS = ['PIXLIMF1', 'PIXLIMF2', 'PIXLIMR1', 'PIXLIMR2', 'PIXLIMM1', 'PIXLIMM2']
WAVLIM_KEYS = ['WAVLIMF1', 'WAVLIMF2', 'WAVLIMR1', 'WAVLIMR2', 'WAVLIMM1', 'WAVLIMM2']


class FluxCalibration(Corrector):
    """A Node that calibrates absolute flux."""

    def __init__(self, sensitivity, datamodel=None, dtype='float32'):

        self.sensp = sensitivity['primary']
        self.pixlims = {key: (self.sensp.header[key] - 1) for key in PIXLIM_KEYS}

        calibid = datamodel.get_imgid(sensitivity)
        self.target_unit = self.sensp.header['TUNIT']
        super(FluxCalibration, self).__init__(datamodel=datamodel,
                                              calibid=calibid,
                                              dtype=dtype)

    def run(self, img):
        imgid = self.get_imgid(img)
        _logger.debug('calibrate flux in image %s', imgid)
        hdr = img['primary'].header
        exptime = hdr['EXPTIME']
        img['primary'].data /= exptime
        newdata = numpy.zeros_like(img['primary'].data)
        limr1 = self.pixlims['PIXLIMR1']
        limr2 = self.pixlims['PIXLIMR2']
        validr = slice(limr1, limr2 + 1, 1)
        # with numpy.errstate(invalid='ignore', divide='ignore'):
        newdata[:, validr] = img['primary'].data[:, validr] / self.sensp.data[validr]
        img['primary'].data = newdata
        self.header_update(hdr, imgid)

        return img

    def header_update(self, hdr, imgid):
        hdr['BUNIT'] = self.target_unit
        hdr['NUM-STD'] = self.calibid
        hdr['NUM-FLUX'] = self.calibid

        update_flux_limits(hdr, self.pixlims, ref=0)

        hdr['history'] = 'Flux calibration in {}'.format(imgid)
        hdr['history'] = 'Flux calibration time {}'.format(datetime.datetime.utcnow().isoformat())


def update_flux_limits(header, pixlims, wcs=None, ref=1):
    """Update keywords used for flux limits"""

    pixlim_coms = ["Start of valid flux calibration",
                   "End of valid flux calibration",
                   'Start of region with at least one fiber',
                   'End of region with at least one fiber',
                   'Start of region with all fibers',
                   'End of region with all fibers',
                   ]

    if ref not in [0, 1]:
        raise ValueError("ref must be 0 or 1")

    off = (ref + 1) % 2

    if wcs is None:
        wcs = astropy.wcs.WCS(header)

    for key, com in zip(PIXLIM_KEYS, pixlim_coms):
        header[key] = (pixlims[key] + off, com)

    r1 = numpy.array([pixlims[key] for key in PIXLIM_KEYS])
    r2 = numpy.zeros_like(r1)
    lm = numpy.array([r1, r2])
    # Values are 0-based
    wavelen_ = wcs.all_pix2world(lm.T, ref)
    wavelen = wavelen_[:, 0]

    for idx, (key, com) in enumerate(zip(WAVLIM_KEYS, pixlim_coms)):
        header[key] = (wavelen[idx], com)

    return header


#
# Copyright 2016-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import numpy
import scipy.interpolate as ii
import astropy.io.fits as fits
import astropy.wcs
from numina.simulation.efficiency import Efficiency


class InterpolFile(object):

    def __init__(self, fname, fill_value=0.0, factor=1.0):
        rawdata = numpy.loadtxt(fname)
        self._interp = ii.interp1d(rawdata[:,0] / 1e4, rawdata[:,1],
                                   bounds_error=False, fill_value=fill_value)
        self.factor = factor

    def __call__(self, wl):
        return self.factor * self._interp(wl)


class InterpolFitsUVES(object):
    """Interpolate spectrum in UVES format.

    This is the format of the sky spectrum file
    """
    def __init__(self, fname, fill_value=0.0):
        with fits.open(fname) as hdul:
            rawdata = hdul[0].data
            wcs = astropy.wcs.WCS(hdul[0].header)
            # WL is in Angstroms -> to microns
            all_wl = wcs.all_pix2world(range(hdul[0].data.size), 0)

        self._interp = ii.interp1d(all_wl[0] / 1e4, numpy.abs(rawdata),
                                   bounds_error=False,
                                   fill_value=fill_value)

    def __call__(self, wl):
        return self._interp(wl)


class EfficiencyFile(Efficiency):

    def __init__(self, fname, fill_value=0.0, factor=1.0):
        self.interpf = InterpolFile(fname, fill_value=fill_value, factor=factor)

    def response(self, wl):
        return self.interpf(wl)



#
# Copyright 2011-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import logging
import multiprocessing as mp
from astropy.io import fits

import numpy as np
from numina.processing import Corrector

_logger = logging.getLogger('numina.processing')


class WeightsCorrector(Corrector):
    '''A Node that corrects from twilight.'''

    def __init__(self, master_weights, datamodel=None, mark=True,
                 tagger=None, dtype='float32'):
        #tagger = TagFits('NUM-MFF', 'MEGARA master_weights correction')

        super(WeightsCorrector, self).__init__(datamodel=datamodel, dtype=dtype)

        self.master_weights = master_weights
        self.processes = mp.cpu_count() - 2
        self.SIZE = 4096

    def decompress(self):
        '''
        :param tar_name: <str> name of the tar file
        :return: None
        '''

        name = self.master_weights.fileobj.name.split('.tar')[0]

        aux = self.master_weights.extractall(name + '/')
        return name

    def run(self, img):
        '''
        :param img: <numpy.array> reduced image
        :param tar_file: <pointer> tar file to be extracted
        :return: list of extracted weights
        '''
        img = img[0].data
        _logger.debug('correct from weights in image ')

        path = self.decompress()

        _logger.info('decompress done')
        _logger.info('Starting: _load_files_paralell')
        pool = mp.Pool(processes=self.processes)
        results = [pool.apply_async(_load_files_paralell,
                                    args=(ite, path)) for ite in
                   range(self.SIZE)]
        results = [p.get() for p in results]
        # return results

        _logger.info('Starting: extract_w_paralell')

        pool2 = mp.Pool(processes=self.processes)
        extracted_w = [pool2.apply_async(extract_w_paralell,
                                         args=(img[:, ite], results[ite])) for
                       ite in range(self.SIZE)]
        extracted_w = [p.get() for p in extracted_w]

        _logger.info('extracted')

        hdu = fits.PrimaryHDU(np.array(extracted_w).T)

        return fits.HDUList([hdu])


def extract_w_paralell(img, mlist):
    '''
    :param img: <fits>
    :param mlist: <list> one element of the csr_matrix
    :return: <ndarray> result of lsqr
    '''
    from scipy.sparse.linalg import lsqr
    x = lsqr(mlist, img)
    return x[0]


def _load_files_paralell(col, path):
    '''
    :param col: <str,int> name of the fits file. It is a counter
    :param path: <str> path where *.npz are
    :return: csr_matrix
    '''
    from scipy.sparse import csr_matrix

    filename = '%s/%s.npz' % (path, col)
    loader = np.load(filename)
    return csr_matrix(
        (loader['data'], loader['indices'], loader['indptr']),
        shape=loader['shape'])



#
# Copyright 2015-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Tests for the bpm mode recipe module."""

from numina.core import DataFrame, ObservationResult
import astropy.io.fits as fits
from numina.core.requirements import ObservationResultRequirement


from megaradrp.core.recipe import MegaraBaseRecipe
from megaradrp.requirements import MasterBiasRequirement, MasterBPMRequirement

from numina.array import combine

from megaradrp.processing.combine import basic_processing_with_combination

from megaradrp.recipes.calibration.tests.test_bpm_common import crear_archivos
from megaradrp.instrument.loader import build_instrument_config, Loader


class DerivedRecipe(MegaraBaseRecipe):
    obresult = ObservationResultRequirement()
    master_bias = MasterBiasRequirement()
    master_bpm = MasterBPMRequirement()

    def __init__(self, directorio):
        self.directorio = directorio
        super(DerivedRecipe, self).__init__()

    def run(self, rinput):
        import copy

        N = len(rinput.obresult.frames)
        obresult1 = copy.copy(rinput.obresult)
        obresult1.frames = rinput.obresult.frames[:N]

        flow1 = self.init_filters(rinput, rinput.obresult.configuration)
        img = basic_processing_with_combination(rinput, flow1, method=combine.median)
        hdr = img[0].header
        self.set_base_headers(hdr)

        reduced1 = img

        fits.writeto(self.directorio + '/reduced_flat.fits', reduced1[0].data,
                     overwrite=True)

        fits.writeto(self.directorio + '/reduced_flat_bpm.fits',
                     reduced1[0].data, overwrite=True)

        return self.create_result()


def test_bpm_corrector():
    import shutil
    from tempfile import mkdtemp

    directorio = mkdtemp()
    names = crear_archivos(directorio, number=4)

    ob = ObservationResult()
    ob.instrument = 'MEGARA'
    ob.mode = 'MegaraBiasImage'
    ob.configuration = build_instrument_config('4fd05b24-2ed9-457b-b563-a3c618bb1d4c', loader=Loader())
    ob.frames = [DataFrame(filename=open(nombre).name) for nombre in names]

    recipe = DerivedRecipe(directorio)
    ri = recipe.create_input(obresult=ob, master_bias=DataFrame(
        filename=open(directorio + '/master_bias_data0.fits').name),
                             master_bpm=DataFrame(filename=open(
                                 directorio + '/master_bpm.fits').name))

    recipe.run(ri)
    shutil.rmtree(directorio)

if __name__ == "__main__":
    test_bpm_corrector()



#
# Copyright 2011-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Base scientific recipe for MEGARA"""


from numina.core import Parameter
from numina.core.requirements import ObservationResultRequirement
from numina.util.flow import SerialFlow
from numina.array import combine
from numina.frame.utils import copy_img

from megaradrp.core.recipe import MegaraBaseRecipe
import megaradrp.requirements as reqs
from megaradrp.processing.combine import basic_processing_with_combination

from megaradrp.processing.aperture import ApertureExtractor
from megaradrp.processing.wavecalibration import WavelengthCalibrator
from megaradrp.processing.fiberflat import FlipLR, FiberFlatCorrector
from megaradrp.processing.twilight import TwilightCorrector
from megaradrp.processing.extractobj import compute_centroid, compute_dar
from megaradrp.processing.extractobj import subtract_sky


class ImageRecipe(MegaraBaseRecipe):
    """Base Image."""

    # Requirements  
    obresult = ObservationResultRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_bpm = reqs.MasterBPMRequirement()
    master_slitflat = reqs.MasterSlitFlatRequirement()
    master_wlcalib = reqs.WavelengthCalibrationRequirement()
    master_fiberflat = reqs.MasterFiberFlatRequirement()
    master_twilight = reqs.MasterTwilightRequirement()
    master_traces = reqs.MasterAperturesRequirement()
    extraction_offset = Parameter([0.0], 'Offset traces for extraction', accept_scalar=True)
    ignored_sky_bundles = Parameter([], 'Ignore these sky bundles')
    master_sensitivity = reqs.SensitivityRequirement()
    reference_extinction = reqs.ReferenceExtinction()
    relative_threshold = Parameter(0.3, 'Threshold for peak detection')

    def base_run(self, rinput):

        # 2D reduction
        flow1 = self.init_filters(rinput, rinput.obresult.configuration)
        img = basic_processing_with_combination(rinput, flow1, method=combine.median)
        hdr = img[0].header
        self.set_base_headers(hdr)

        self.save_intermediate_img(img, 'reduced_image.fits')

        reduced2d = copy_img(img)

        # 1D, extraction, Wl calibration, Flat fielding
        reduced_rss = self.run_reduction_1d(img,
            rinput.master_traces, rinput.master_wlcalib,
            rinput.master_fiberflat, rinput.master_twilight,
            offset=rinput.extraction_offset
        )
        self.save_intermediate_img(reduced_rss, 'reduced_rss.fits')

        return reduced2d, reduced_rss

    def run_reduction_1d(self, img, tracemap, wlcalib, fiberflat, twflat=None, offset=None):
        # 1D, extraction, Wl calibration, Flat fielding
        correctors = []
        correctors.append(ApertureExtractor(tracemap, self.datamodel, offset=offset))
        correctors.append(FlipLR())
        correctors.append(WavelengthCalibrator(wlcalib, self.datamodel))
        correctors.append(FiberFlatCorrector(fiberflat.open(), self.datamodel))

        if twflat:
            correctors.append(TwilightCorrector(twflat.open(), self.datamodel))

        flow2 = SerialFlow(correctors)

        reduced_rss =  flow2(img)
        return reduced_rss

    def run_sky_subtraction(self, img, ignored_sky_bundles=None):
        return subtract_sky(img,
                            self.datamodel,
                            ignored_sky_bundles=ignored_sky_bundles,
                            logger=self.logger
                            )

    def compute_dar(self, img):
        import numpy.polynomial.polynomial as pol

        wl, xdar, ydar = compute_dar(img, self.datamodel, logger=self.logger)
        print('DAR, x:', pol.polyfit(wl, xdar, deg=3))
        print('DAR: y:', pol.polyfit(wl, ydar, deg=3))

    def centroid(self, rssdata, fiberconf, c1, c2, point):
        return compute_centroid(rssdata, fiberconf, c1, c2, point, logger=self.logger)



#
# Copyright 2015-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import shutil
from tempfile import mkdtemp

import astropy.io.fits as fits
import numpy
from numina.core import DataFrame, ObservationResult

from megaradrp.recipes.calibration.dark import DarkRecipe
from megaradrp.simulation.factory import MegaraImageFactory
from megaradrp.simulation.detector import ReadParams, MegaraDetectorSat
from megaradrp.simulation.actions import simulate_dark_fits
from megaradrp.instrument.loader import build_instrument_config, Loader


def test_dark():

    numpy.random.seed(422992983)

    PSCAN = 50
    DSHAPE = (2056 * 2, 2048 * 2)
    OSCAN = 50

    # ron = 2.0
    ron = 0.001
    gain = 1.0
    bias = 1000.0
    dark = 3.0 # In 1 hour
    exptime = 3600.0
    dark_s = dark / exptime
    qe = 0.8 * numpy.ones(DSHAPE)
    config_uuid = '4fd05b24-2ed9-457b-b563-a3c618bb1d4c'
    temporary_path = mkdtemp()
    fits.writeto('%s/eq.fits' % temporary_path, qe, overwrite=True)

    readpars1 = ReadParams(gain=gain, ron=ron, bias=bias)
    readpars2 = ReadParams(gain=gain, ron=ron, bias=bias)

    detector = MegaraDetectorSat('megara_test_detector', DSHAPE, OSCAN, PSCAN, qe=qe,
                                 dark=dark_s,
                                 readpars1=readpars1, readpars2=readpars2,
                                 bins='11')

    number = 3
    factory = MegaraImageFactory()
    fs = simulate_dark_fits(factory, detector, exposure=3600, repeat=number)

    for idx, aux in enumerate(fs):
        aux.writeto('%s/dark_%s.fits' % (temporary_path, idx), overwrite=True)

    header = fits.Header()
    header['DATE-OBS'] = '2017-11-09T11:00:00.0'
    master_bias_data = numpy.zeros(DSHAPE)
    master_bias_hdul = fits.HDUList(fits.PrimaryHDU(
        master_bias_data, header=header)
    )
    #master_bias_data = master_bias.master_bias.frame[0].data

    ob = ObservationResult()
    ob.instrument = 'MEGARA'
    ob.mode = 'MegaraDarkImage'
    ob.configuration = build_instrument_config(config_uuid, loader=Loader())

    names = []
    for aux in range(number):
        names.append('%s/dark_%s.fits' % (temporary_path, aux))
    ob.frames = [DataFrame(filename=open(nombre).name) for nombre in names]

    recipe = DarkRecipe()
    ri = recipe.create_input(
        obresult=ob,
        master_bias=DataFrame(frame=master_bias_hdul),
    )
    aux = recipe.run(ri)

    mean_dark_value = aux.master_dark.frame[0].data.mean()

    shutil.rmtree(temporary_path)

    assert numpy.allclose(mean_dark_value, dark, atol=0, rtol=1e-1)


if __name__ == "__main__":
    test_dark()



#
# Copyright 2011-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Load image correctos according to present calibrations"""


import logging


import numina.util.node as node
import numina.processing as proc

from megaradrp.processing.trimover import OverscanCorrector, TrimImage
from megaradrp.processing.trimover import GainCorrector
from megaradrp.processing.slitflat import SlitFlatCorrector


_logger = logging.getLogger(__name__)


def get_corrector_bpm(rinput, meta, ins, datamodel):
    bpm_info = meta.get('master_bpm')
    if bpm_info is not None:
        with rinput.master_bpm.open() as hdul:
            _logger.info('loading BPM')
            mbpm = hdul[0].data
            calibid = datamodel.get_imgid(hdul)
            _logger.debug('BPM image: %s', calibid)
            bpm_corrector = proc.BadPixelCorrector(
                mbpm,
                datamodel=datamodel,
                calibid=calibid
            )
    else:
        _logger.info('BPM not provided, ignored')
        bpm_corrector = node.IdNode()

    return bpm_corrector


def get_corrector_super(rinput, meta, key, correctorclass, datamodel):
    info = meta.get(key)
    req = getattr(rinput, key)
    if req is not None:
        with req.open() as hdul:
            _logger.info('loading %s', key)
            _logger.debug('%s info: %s', key, info)
            datac = hdul['primary'].data
            calibid = datamodel.get_imgid(hdul)
            corrector = correctorclass(datac, datamodel=datamodel,
                                       calibid=calibid)
    else:
        _logger.info('%s not provided, ignored', key)
        corrector = node.IdNode()

    return corrector


def get_corrector_bias(rinput, meta, ins, datamodel):
    key = 'master_bias'
    correctorclass = proc.BiasCorrector
    return get_corrector_super(rinput, meta, key, correctorclass, datamodel)


def get_corrector_dark(rinput, meta, ins, datamodel):
    key = 'master_dark'
    correctorclass = proc.DarkCorrector
    return get_corrector_super(rinput, meta, key, correctorclass, datamodel)


def get_corrector_slit_flat(rinput, meta, ins, datamodel):
    key = 'master_slitflat'
    info = meta.get(key)
    if info is not None:
        req = getattr(rinput, key)
        with req.open() as hdul:
            _logger.info('loading slit flat')
            _logger.debug('%s image: %s', key, info)
            mbpm = hdul[0].data
            calibid = datamodel.get_imgid(hdul)
            corrector = SlitFlatCorrector(mbpm, datamodel, calibid=calibid)
    else:
        _logger.info('%s not provided, ignored', key)
        corrector = node.IdNode()

    return corrector


def get_corrector_overscan(rinput, meta, ins, datamodel):
    detconf = ins.get('detector.scan')
    return OverscanCorrector(
        detconf,
        datamodel=datamodel,
        calibid=ins.components['detector'].uuid
    )


def get_corrector_trimming(rinput, meta, ins, datamodel):
    detconf = ins.get('detector.scan')
    return TrimImage(
        detconf,
        datamodel=datamodel,
        calibid=ins.components['detector'].uuid
    )


def get_corrector_gain(rinput, meta, ins, datamodel):
    """Correct from gain"""
    return GainCorrector(
        detconf=ins.get('detector.scan'),
        datamodel=datamodel,
        calibid=ins.components['detector'].uuid
    )



from megaradrp.processing.trimover import OverscanCorrector
from numina.core import DataFrame
import shutil
from tempfile import mkdtemp
import numpy as np
import astropy.io.fits as fits

def test_OverscanCorrector():

    detconf = {
        'trim1': [[0, 2056], [50, 4146]],
        'trim2': [[2156, 4212], [50, 4146]],
        'bng': [1, 1],
        'overscan1': [[0, 2056], [4149, 4196]],
        'overscan2': [[2156, 4212], [0, 50]],
        'prescan1': [[0, 2056], [0, 50]],
        'prescan2': [[2156, 4212], [4145, 4196]],
        'middle1': [[2056, 2106], [50, 4146]],
        'middle2': [[2106, 2156], [50, 4146]]
    }

    oc = OverscanCorrector(detconf)
    assert oc.trim1 == (slice(0, 2056, None),slice(50, 4146, None))
    assert oc.pcol1 == (slice(0, 2056, None),slice(0, 50, None))
    assert oc.ocol1 == (slice(0, 2056, None),slice(4149, 4196, None))
    assert oc.orow1 == (slice(2056, 2106, None),slice(50, 4146, None))

    assert oc.trim2 == (slice(2156, 4212, None),slice(50, 4146, None))
    assert oc.pcol2 == (slice(2156, 4212, None),slice(4145, 4196, None))
    assert oc.ocol2 == (slice(2156, 4212, None),slice(0, 50, None))
    assert oc.orow2 == (slice(2106, 2156, None),slice(50, 4146, None))


if __name__ == "__main__":
    test_OverscanCorrector()



#
# Copyright 2016-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import six

from numina.simulation.wheel import Wheel
from numina.simulation.optics import Open, Stop, Filter


class MegaraShutter(Wheel):
    def __init__(self, parent=None):
        super(MegaraShutter, self).__init__(capacity=3, name='Shutter', parent=parent)
        self.put_in_pos(Stop(name='STOP'), 0) # FIXME
        self.put_in_pos(Open(name='OPEN'), 1) # FIXME
        # sorting order filter
        self.put_in_pos(Filter(transmission=None, name='FILTER'), 2) # FIXME
        self.move_to(1) # Open by default

        # MEGARA shutter has three positions:
        # open
        # closed
        # filter (sort order)

    def configure(self, value):
        # Let's see what is value:
        # a string
        if isinstance(value, six.string_types):
            val = value.lower()
            if val == 'open':
                val = 1
            elif val == 'closed':
                val = 0
            elif val == 'filter':
                val = 2
            else:
                raise ValueError('Not allowed value %s', value)
        elif isinstance(value, (int, long)):
            val = value
        else:
            raise TypeError('Not allowed type %s', type(value))

        # Move to value
        self.move_to(val)

    def open(self):
        self.move_to(1)

    def filter(self):
        self.move_to(2)

    def close(self):
        self.move_to(0)




import megaradrp.processing.extractobj as eobj


def test_cover_1():

    arr = [1, 1, 1, 1, 1, 1]
    res = eobj.coverage_det(arr)
    assert res == slice(0, 7)

    arr = [0, 0, 0, 0, 0, 0]
    res = eobj.coverage_det(arr)
    assert res == slice(0, 0)


def test_cover_2():

    arr = [0, 1, 1, 1, 0, 0]
    res = eobj.coverage_det(arr)
    assert res == slice(1, 4, None)

    arr = [0, 1, 1, 1, 1, 0]
    res = eobj.coverage_det(arr)
    assert res == slice(1, 5)

    arr = [0, 1, 1, 1, 1, 1]
    res = eobj.coverage_det(arr)
    assert res == slice(1, 7)

    arr = [1, 1, 1, 1, 1, 0]
    res = eobj.coverage_det(arr)
    assert res == slice(0, 5)



#
# Copyright 2011-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""LCB Direct Image Recipe for Megara"""

import numpy
from scipy.interpolate import interp1d
import astropy.wcs

from numina.core import Product

from megaradrp.recipes.scientific.base import ImageRecipe
from megaradrp.types import ProcessedRSS, ProcessedFrame
from megaradrp.processing.fluxcalib import FluxCalibration


class LCBImageRecipe(ImageRecipe):
    """Process LCB images.

    This recipe processes a set of images
    obtained in **LCB image** mode and returns
    the sky subtracted RSS.

    See Also
    --------
    megaradrp.recipes.scientific.mos.MOSImageRecipe

    Notes
    -----
    Images provided by `obresult` are trimmed and corrected
    from overscan, bad pixel mask (if `master_bpm` is not None),
    bias, dark current (if `master_dark` is not None) and
    slit-flat (if `master_slitflat` is not None).

    Images thus corrected are the stacked using the median.
    The result of the combination is saved as an intermediate result, named
    'reduced_image.fits'. This combined image is also returned in the field
    `reduced_image` of the recipe result.

    The apertures in the 2D image are extracted, using the information in
    `master_traces` and resampled according to the wavelength calibration in
    `master_wlcalib`. Then is divided by the `master_fiberflat`.
    The resulting RSS is saved as an intermediate
    result named 'reduced_rss.fits'. This RSS is also returned in the field
    `reduced_rss` of the recipe result.

    The sky is subtracted by combining the the fibers marked as `SKY`
    in the fibers configuration. The RSS with sky subtracted is returned in the
    field `final_rss` of the recipe result.

    If a `master_sensitivity` is provided (optional), RSS products will be
    flux calibrated. If `reference_extinction` is provided (optional), `final_rss`
    and `reduced_rss` will be extinction corrected. Notice that
    `sky_rss` is not corrected from extinction.

    """

    reduced_image = Product(ProcessedFrame)
    final_rss = Product(ProcessedRSS)
    reduced_rss = Product(ProcessedRSS)
    sky_rss = Product(ProcessedRSS)

    def run(self, rinput):

        self.logger.info('starting LCB reduction')

        reduced2d, rss_data = super(LCBImageRecipe,self).base_run(rinput)

        self.logger.info('start sky subtraction')
        isb = rinput.ignored_sky_bundles
        if isb:
            self.logger.info('sky bundles ignored: %s', isb)
        final, origin, sky = self.run_sky_subtraction(rss_data, ignored_sky_bundles=isb)
        self.logger.info('end sky subtraction')
        # Flux calibration
        if rinput.master_sensitivity is not None:
            self.logger.info('start flux calibration')
            node = FluxCalibration(rinput.master_sensitivity.open(), self.datamodel)
            final = node(final)
            origin = node(origin)
            sky = node(sky)
            self.logger.info('end flux calibration')
        else:
            self.logger.info('no flux calibration')

        # Extinction calibration
        if rinput.reference_extinction is not None:
            self.logger.info('start extinction correction')
            extinc_interp = interp1d(rinput.reference_extinction[:, 0],
                                     rinput.reference_extinction[:, 1])

            wlcalib = astropy.wcs.WCS(final[0].header)
            pixrange = numpy.arange(final[0].data.shape[1])
            yrange = pixrange * 0
            calcwl = numpy.array([pixrange, yrange]).T
            wavelen = wlcalib.all_pix2world(calcwl, 0.0)[:,0]
            airmass = final[0].header['AIRMASS']

            extinc_corr = numpy.power(10.0, 0.4 * extinc_interp(wavelen) * airmass)

            final[0].data *= extinc_corr
            origin[0].data *= extinc_corr
            # sky is not corrected from extinction
            self.logger.info('end extinction correction')
        else:
            self.logger.info('no extinction correction')

        self.logger.info('end LCB reduction')

        return self.create_result(
            reduced_image=reduced2d,
            final_rss=final,
            reduced_rss=origin,
            sky_rss=sky
        )



#
# Copyright 2011-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#


"""Match and identify fibers"""

import itertools
import collections


from six.moves import filter as ifilter


FIBER_PEAK, FIBER_DEAD = (0, 1)
FOUND_PEAK, FOUND_VALLEY, EXPECT_VALLEY = (0, 1, 2)


FiberModelElement = collections.namedtuple('FiberModelElement', ['fibid', 'mode'])


def generate_box_model(nfibers, start=1,
                       missing_relids=None,
                       skip_fibids=None
                       ):
    """Generate a model of the expected peaks in a box"""

    if skip_fibids is None:
        skip_fibids = []

    if missing_relids is None:
        missing_relids = []

    iter1 = itertools.count(start)
    iter2 = ifilter(lambda x:x not in skip_fibids, iter1)
    iter3 = itertools.islice(iter2, nfibers)

    result = []
    for idx, fibid in enumerate(iter3, 1):
        key = FIBER_PEAK
        if idx in missing_relids:
            key = FIBER_DEAD
        tok = FiberModelElement(fibid=fibid, mode=key)
        result.append(tok)
    return result


def count_peaks(peaks, tol=1.2, distance=6.0, start=1, max_scale_jump=3):
    """Count the peaks and valleys of an array"""

    expected_distance = distance

    scale = 1
    pidx = 0
    pid = start

    if len(peaks) == 0:
        raise ValueError('no peaks to count')

    p1, rest = peaks[0], peaks[1:]
    # pref = p1
    values = [(pid, p1, FOUND_PEAK, 0)]

    while len(rest) > 0:
        # print('im peak:', p1, ' next peak should be around:', p1 + scale * expected_distance)
        p2 = rest[0]
        dist = abs(p1 - p2)
        while True:
            sed = scale * expected_distance
            # print('next peak is:', p2, 'distance from p1 is', dist)
            # print('expected distance is:', sed)
            pid += 1

            if abs(dist - sed) < scale * tol:
                # print('p2 is within expected distance with tol:', tol)
                # print('p2 is next peak, with scale', scale)
                pidx += 1
                values.append((pid, p2, FOUND_PEAK, pidx))
                scale = 1
                break
            else:
                # print('p2 is not within expected distance with tol:', tol)
                pex = p1 + sed
                values.append((pid, pex, FOUND_VALLEY, None))
                scale += 1
                # print('increase scale to:', scale)
                if scale > max_scale_jump:
                    # print('moving to far')
                    raise ValueError('moving too far apart')

        p1, rest = rest[0], rest[1:]
    return values


# matching
def valid_match(model, values):
    """Match a model with found peaks"""
    for m, v in zip(model, values):
        field_m = m.mode
        field_v = v[2]
        if field_m == FIBER_DEAD and field_v == FOUND_PEAK:
            # print(m, v, 'must be empty')
            return False
    return True


def complete_solutions(model, values, borders, scale=6.0):
    """Try different solutions to match the missing peaks"""

    border1 = borders[0]
    border2 = borders[1]

    missing = len(model) - len(values)

    # last_peak = values[-1][1]
    last_peak = values[-1]
    first_peak = values[0]
    # distance to borders
    d1 = abs(first_peak[1] - border1)
    d2 = abs(last_peak[1] - border2)

    solutions = []

    # print('XXX missing peaks', missing)
    for pl in range(missing + 1):
        pr = missing - pl
        # print('XXXX', pl, pr)
        # Complete peaks
        # preapend pl peaks

        pre = []
        post = []
        # print('XXXX space per peak added', d1 / (pl + 1.0), d2 / (pr + 1.0))
        for peak in range(pl):
            idx = -peak - 1
            dist = scale * idx
            tok = (idx, dist, EXPECT_VALLEY, None)
            pre.append(tok)

        for peak in range(pr):
            idx = (len(values) + peak) + 1
            dist = last_peak[1] + scale * (peak + 1)
            tok = (idx, dist, EXPECT_VALLEY, None)
            post.append(tok)

        if valid_match(model, itertools.chain(pre, values, post)):
            # print('computing score')
            # print(border1, border2)
            ratios = [d1 / (pl +1 ), d2 / (pr + 1)]
            # print('ratios:', ratios)
            exp_dist = scale
            score = sum([(c - exp_dist) ** 2 for c in ratios])
            # print('score', score)
            solution = (score, (pre, post))
            solutions.append(solution)

    # print('------------')
    return solutions


def iter_best_solution(model, values, solutions):

    # This works because score is the first...
    best_solution = min(solutions)
    minscore, added_peaks = best_solution

    pre, post = added_peaks
    comp_sol = itertools.chain(pre, values, post)
    for xx, yy in zip(comp_sol, model):
        fibid = yy[0]
        match = xx[3]
        yield (fibid, match)


#
# Copyright 2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Tests for the trace map mode recipe module."""

import pytest

from numina.user.cli import main

from megaradrp.loader import load_drp


BASE_URL = 'http://guaix.fis.ucm.es/~spr/megara_test/'


def run_recipe():

    main(['run', 'obsrun.yaml', '-r', 'control.yaml'])


@pytest.mark.remote
@pytest.mark.usefixtures("numinatpldir")
def test_mode_trace_map_set0(drpmocker):

    drpmocker.add_drp('MEGARA', load_drp)

    run_recipe()



#
# Copyright 2015-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Tests for the bpm mode recipe module."""

import astropy.io.fits as fits
import numpy as np
from numina.core import DataFrame, ObservationResult

from megaradrp.instrument.loader import build_instrument_config, Loader

def generate_bias(detector, number, temporary_path):
    from megaradrp.simulation.actions import simulate_bias
    from megaradrp.recipes.calibration.bias import BiasRecipe

    config_uuid = '4fd05b24-2ed9-457b-b563-a3c618bb1d4c'
    fs = [simulate_bias(detector) for i in range(number)]
    header = fits.Header()
    header['DATE-OBS'] = '2017-11-09T11:00:00.0'
    for aux in range(len(fs)):
        fits.writeto('%s/bias_%s.fits' % (temporary_path, aux), fs[aux],
                     header=header,
                     overwrite=True)

    fs = ["%s/bias_%s.fits" % (temporary_path, i) for i in range(number)]

    ob = ObservationResult()
    ob.instrument = 'MEGARA'
    ob.mode = 'bias_image'
    ob.configuration = build_instrument_config(config_uuid, loader=Loader())
    ob.frames = [DataFrame(filename=f) for f in fs]

    recipe = BiasRecipe()
    ri = recipe.create_input(obresult=ob)
    return recipe.run(ri)


def crear_archivos(temporary_path, number=5):
    from megaradrp.simulation.actions import simulate_flat
    from megaradrp.simulation.detector import ReadParams, MegaraDetectorSat
    from megaradrp.recipes.calibration.bpm import BadPixelsMaskRecipe

    config_uuid = '4fd05b24-2ed9-457b-b563-a3c618bb1d4c'
    PSCAN = 50
    DSHAPE = (2056 * 2, 2048 * 2)
    OSCAN = 50
    ron = 2.0
    gain = 1.0
    bias = 1000.0

    qe = 0.8 * np.ones(DSHAPE)
    qe[0:15, 0:170] = 0.0

    readpars1 = ReadParams(gain=gain, ron=ron, bias=bias)
    readpars2 = ReadParams(gain=gain, ron=ron, bias=bias)

    detector = MegaraDetectorSat('megara_test_detector', DSHAPE, OSCAN, PSCAN,
                                 qe=qe,
                                 dark=(3.0 / 3600.0),
                                 readpars1=readpars1, readpars2=readpars2,
                                 bins='11')

    source2 = 1.0

    fs = [simulate_flat(detector, exposure=1.0, source=5000 * source2) for i in
          range(number)]

    header = fits.Header()
    header['DATE-OBS'] = '2017-11-09T11:00:00.0'
    for aux in range(len(fs)):
        fits.writeto('%s/flat_%s.fits' % (temporary_path, aux), fs[aux],
                     header=header,
                     overwrite=True)

    result = generate_bias(detector, number, temporary_path)
    result.master_bias.frame.writeto(
        '%s/master_bias_data0.fits' % temporary_path,
        overwrite=True)  # Master Bias

    ob = ObservationResult()
    ob.instrument = 'MEGARA'
    ob.mode = 'bias_image'
    ob.configuration = build_instrument_config(config_uuid, loader=Loader())

    names = []
    for aux in range(number):
        names.append('%s/flat_%s.fits' % (temporary_path, aux))
    ob.frames = [DataFrame(filename=open(nombre).name) for nombre in names]

    recipe = BadPixelsMaskRecipe()
    ri = recipe.create_input(obresult=ob, master_bias=DataFrame(
        filename=open(temporary_path + '/master_bias_data0.fits').name))
    aux = recipe.run(ri)
    aux.master_bpm.frame.writeto('%s/master_bpm.fits' % temporary_path, overwrite=True)

    return names



#
# Copyright 2011-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Products of the Megara Pipeline"""


import numina.types.structured as structured
import numina.core.tagexpr as tagexpr

import megaradrp.datamodel


class BaseStructuredCalibration(structured.BaseStructuredCalibration):
    def __init__(self, instrument='unknown'):
        datamodel = megaradrp.datamodel.MegaraDataModel()
        super(BaseStructuredCalibration, self).__init__(instrument, datamodel)
        my_tag_table = self.datamodel.query_attrs
        objtags = [my_tag_table[t] for t in self.tag_names()]
        self.query_expr = tagexpr.query_expr_from_attr(objtags)
        self.names_t = self.query_expr.tags()
        self.names_f = self.query_expr.fields()

        self.total_fibers = 0
        self.missing_fibers = []
        self.error_fitting = []

    def __getstate__(self):
        st = super(BaseStructuredCalibration, self).__getstate__()

        keys = ["total_fibers", "missing_fibers", "error_fitting"]

        for key in keys:
            st[key] = self.__dict__[key]

        return st



#
# Copyright 2015-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Tests for the bpm mode recipe module."""
import shutil
from tempfile import mkdtemp

import astropy.io.fits as fits
import numpy as np

from numina.core import DataFrame, ObservationResult

from megaradrp.recipes.calibration.bpm import BadPixelsMaskRecipe
from megaradrp.recipes.calibration.tests.test_bpm_common import generate_bias
from megaradrp.simulation.detector import ReadParams, MegaraDetectorSat
from megaradrp.simulation.actions import simulate_flat
from megaradrp.instrument.loader import build_instrument_config, Loader


# @pytest.mark.remote
def test_bpm():
    number = 5
    PSCAN = 50
    DSHAPE = (2056 * 2, 2048 * 2)
    OSCAN = 50

    ron = 2.0
    gain = 1.0
    bias = 1000.0

    qe = 0.8 * np.ones(DSHAPE)
    qe[5:6, 0:170] = 0.0
    config_uuid = '4fd05b24-2ed9-457b-b563-a3c618bb1d4c'
    temporary_path = mkdtemp()

    fits.writeto('%s/eq.fits' % temporary_path, qe, overwrite=True)

    readpars1 = ReadParams(gain=gain, ron=ron, bias=bias)
    readpars2 = ReadParams(gain=gain, ron=ron, bias=bias)

    detector = MegaraDetectorSat('megara_test_detector', DSHAPE, OSCAN, PSCAN,
                                 qe=qe,
                                 dark=(3.0 / 3600.0),
                                 readpars1=readpars1, readpars2=readpars2,
                                 bins='11')

    source2 = 1.0

    fs = [simulate_flat(detector, exposure=1.0, source=5000 * source2) for i in
          range(number)]
    fs2 = [simulate_flat(detector, exposure=1.0, source=40000 * source2) for i
           in range(number)]

    header = fits.Header()
    header['DATE-OBS'] = '2017-11-09T11:00:00.0'
    for aux in range(len(fs)):
        fits.writeto('%s/flat_%s.fits' % (temporary_path, aux), fs[aux],
                     header=header,
                     overwrite=True)
        fits.writeto('%s/flat_%s.fits' % (temporary_path, aux + number), fs2[aux],
                     header=header,
                     overwrite=True)

    result = generate_bias(detector, number, temporary_path)
    result.master_bias.frame.writeto(
        '%s/master_bias_data0.fits' % temporary_path,
        overwrite=True
    )

    ob = ObservationResult()
    ob.instrument = 'MEGARA'
    ob.mode = 'MegaraBiasImage'
    ob.configuration = build_instrument_config(config_uuid, loader=Loader())

    names = []
    for aux in range(number * 2):
        names.append('%s/flat_%s.fits' % (temporary_path, aux))
    ob.frames = [DataFrame(filename=open(nombre).name) for nombre in names]

    recipe = BadPixelsMaskRecipe()
    ri = recipe.create_input(obresult=ob, master_bias=DataFrame(
        filename=open(temporary_path + '/master_bias_data0.fits').name))
    aux = recipe.run(ri)
    aux.master_bpm.frame.writeto('%s/master_bpm.fits' % temporary_path, overwrite=True)
    shutil.rmtree(temporary_path)


if __name__ == "__main__":
    test_bpm()



#
# Copyright 2016-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Corrector for wavecalibration"""

import logging
import datetime

import numpy
from numpy.polynomial.polynomial import polyval
from astropy.io import fits
import numina.array.utils as u
from numina.processing import Corrector
from numina.array.interpolation import SteffenInterpolator

from megaradrp.instrument import WLCALIB_PARAMS

_logger = logging.getLogger(__name__)


class WavelengthCalibrator(Corrector):
    """A Node that applies wavelength calibration."""

    def __init__(self, solutionwl, datamodel=None, dtype='float32'):

        super(WavelengthCalibrator, self).__init__(
            datamodel=datamodel,
            calibid=solutionwl.calibid,
            dtype=dtype)

        self.solutionwl = solutionwl

    def run(self, rss):
        imgid = self.get_imgid(rss)
        _logger.debug('wavelength calibration in image %s', imgid)
        _logger.debug('with wavecalib %s', self.calibid)
        _logger.debug('offsets are %s', self.solutionwl.global_offset.coef)

        current_vph = rss[0].header['VPH']
        current_insmode = rss[0].header['INSMODE']

        _logger.debug('Current INSMODE is %s, VPH is %s', current_insmode, current_vph)
        if current_insmode in WLCALIB_PARAMS and current_vph in WLCALIB_PARAMS[current_insmode]:
            wvpar_dict = WLCALIB_PARAMS[current_insmode][current_vph]
            _logger.info('precomputed wl parameters are %s', wvpar_dict)
        else:
            msg = 'insmode {} grism {} is not defined in megaradrp.instrument.WLCALIB_PARAMS'.format(
                current_insmode,
                current_vph
            )
            raise ValueError(msg)

        _logger.debug('Resample RSS')
        final, wcsdata, values = self.resample_rss_flux(rss[0].data, wvpar_dict)

        _logger.debug('Update headers')
        rss_wl = fits.PrimaryHDU(
            data=final.astype(self.dtype),
            header=rss[0].header
        )

        hdr = rss_wl.header
        self.add_wcs(rss_wl.header, wvpar_dict['crval'], wvpar_dict['cdelt'],
                     wvpar_dict['crpix'])

        hdr['NUM-WAV'] = self.calibid
        hdr['history'] = 'Wavelength calibration with {}'.format(self.calibid)
        hdr['history'] = 'Aperture extraction offsets are {}'.format(
            self.solutionwl.global_offset.coef.tolist())
        hdr['history'] = 'Wavelength calibration time {}'.format(datetime.datetime.utcnow().isoformat())

        # Update other HDUs if needed
        rss[0] = rss_wl

        map_data = numpy.zeros(rss[0].shape, dtype='int16')

        fibers_ext = rss['FIBERS']
        fibers_ext_headers = fibers_ext.header

        for fibid, (s1, s2) in values:
            idx = fibid - 1
            map_data[idx, s1:s2+1] = 1
            # Update Fibers
            key = "FIB%03dW1" % fibid
            fibers_ext_headers[key] =  (s1 + 1, "Start of spectral coverage")
            key = "FIB%03dW2" % fibid
            fibers_ext_headers[key] =  (s2 + 1, "End of spectral coverage")

        for fibid in self.solutionwl.error_fitting:
            # Update Fibers
            key = "FIB%03d_V" % fibid
            fibers_ext_headers[key] =  False

        for fibid in self.solutionwl.missing_fibers:
            # Update Fibers
            key = "FIB%03d_V" % fibid
            fibers_ext_headers[key] =  False

        rss_map = fits.ImageHDU(data=map_data, name='WLMAP')

        rss.append(rss_map)
        return rss

    def add_wcs(self, hdr, wlr0, delt, crpix=0.0):
        c_crpix = 'Pixel coordinate of reference point'
        c_cunit = 'Units of coordinate increment and value'
        unit = 'Angstrom'
        c_crval = '[{}] Coordinate value at reference point'.format(unit)
        c_cdelt = '[{}] Coordinate increment at reference point'.format(unit)
        hdr['CRPIX1'] = (crpix, c_crpix)
        hdr['CRVAL1'] = wlr0, c_crval
        hdr['CDELT1'] = delt, c_cdelt
        hdr['CUNIT1'] = unit, c_cunit
        hdr['CTYPE1'] = 'WAVELENGTH'
        hdr['CRPIX2'] = (0.0, c_crpix)
        hdr['CRVAL2'] = 0.0
        hdr['CDELT2'] = 1.0
        hdr['CTYPE2'] = ''
        return hdr

    def resample_rss_flux(self, rss_old, wvpar_dict):

        nfibers = rss_old.shape[0]
        nsamples = rss_old.shape[1]

        npix = wvpar_dict['npix']
        delts = wvpar_dict['cdelt']
        wl_min = wvpar_dict['crval']
        crpix = wvpar_dict['crpix']

        wl_max = wl_min + (npix - crpix) * delts

        new_x = numpy.arange(npix)
        new_wl = wl_min + delts * new_x

        old_x_borders = numpy.arange(-0.5, nsamples)
        old_x_borders += crpix  # following FITS criterium
        # old_wl_borders = polyval(old_x_borders, wcalib.T)

        new_borders = self.map_borders(new_wl)

        accum_flux = numpy.empty((nfibers, nsamples + 1))
        accum_flux[:, 1:] = numpy.cumsum(rss_old, axis=1)
        accum_flux[:, 0] = 0.0
        rss_resampled = numpy.zeros((nfibers, npix))
        values = []

        for fibsol in self.solutionwl.contents:

            fibid = fibsol.fibid
            idx = fibid - 1

            # small correction defined in master_wlcalib_XXX_XX-X.json
            coeff = fibsol.solution.coeff
            offset_wavelength = self.solutionwl.global_offset(fibid)
            coeff[0] -= offset_wavelength

            old_wl_borders = polyval(old_x_borders, coeff)

            s1 = u.coor_to_pix_1d((old_wl_borders[0] - wl_min) / delts)
            s2 = u.coor_to_pix_1d((old_wl_borders[-1] - wl_min) / delts)

            s1 = max(0, min(s1, npix - 1))
            s2 = max(0, min(s2, npix - 1))

            # We need a monotonic interpolator
            # linear would work, we use a cubic interpolator
            interpolator = SteffenInterpolator(
                old_wl_borders,
                accum_flux[idx],
                extrapolate='border'
            )
            fl_borders = interpolator(new_borders)
            rss_resampled[idx] = fl_borders[1:] - fl_borders[:-1]
            values.append((fibid, (s1, s2)))

        return rss_resampled, (wl_min, wl_max, delts), values

    def map_borders(self, wls):
        """Compute borders of pixels for interpolation.

        The border of the pixel is assumed to be midway of the wls
        """
        midpt_wl = 0.5 * (wls[1:] + wls[:-1])
        all_borders = numpy.zeros((wls.shape[0] + 1,))
        all_borders[1:-1] = midpt_wl
        all_borders[0] = 2 * wls[0] - midpt_wl[0]
        all_borders[-1] = 2 * wls[-1] - midpt_wl[-1]
        return all_borders